from django.core.exceptions import ValidationError

__all__ = ('validate_comma_separated',)


def validate_comma_separated(value):
    if not isinstance(value, str):
        raise ValidationError("This field requires a string of comma-separated values.")
    values = value.split(',')
    if any(not item.strip() for item in values):
        raise ValidationError("Each value should be non-empty and comma-separated without extra spaces.")

    return value



# Generated by Django 4.1.10 on 2023-12-26 21:43

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import taggit.managers
import utilities.json


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('extras', '0092_delete_jobresult'),
    ]

    operations = [
        migrations.CreateModel(
            name='NetBoxConfiguration',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('description', models.CharField(blank=True, max_length=200)),
                ('comments', models.TextField(blank=True)),
                ('key', models.CharField(max_length=255, unique=True, validators=[django.core.validators.MinLengthValidator(40)])),
                ('env_name_storage', models.CharField(blank=True, max_length=255, unique=True, validators=[django.core.validators.MinLengthValidator(5)])),
                ('license', models.CharField(blank=True, max_length=255, null=True, validators=[django.core.validators.MinLengthValidator(40), django.core.validators.MaxLengthValidator(93)])),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'verbose_name': 'NetBox Configuration',
                'verbose_name_plural': 'NetBox Configurations',
            },
        ),
        migrations.CreateModel(
            name='NetBoxDBBackup',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('crontab', models.CharField(default='@daily', max_length=255)),
                ('keep_backups', models.PositiveIntegerField(default=1)),
                ('netbox_env', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='db_backups', to='netbox_paas.netboxconfiguration')),
            ],
            options={
                'verbose_name': 'NetBox DB Backup',
                'verbose_name_plural': 'NetBox DB Backups',
            },
        ),
    ]



from netbox.api.routers import NetBoxRouter

from . import views

app_name = "netbox_paas"

router = NetBoxRouter()
router.register("configurations", views.NetBoxConfigurationViewSet)
router.register("db-backups", views.NetBoxDBBackupViewSet)
router.register("plugins", views.NetBoxPluginViewSet, basename="plugins")

urlpatterns = router.urls



import os
from importlib.metadata import metadata

from extras.plugins import PluginConfig

metadata = metadata("netbox_paas")


class NetBoxPaas(PluginConfig):
    name = metadata.get("Name").replace("-", "_")
    verbose_name = metadata.get("Name")
    description = metadata.get("Summary")
    version = metadata.get("Version")
    author = metadata.get("Author")
    author_email = metadata.get("Author-email")
    base_url = "paas"
    min_version = "3.7.0"
    max_version = "3.7.99"
    django_apps = ["elasticapm.contrib.django"]

    def ready(self):
        super().ready()

        # Ensure ENV_NAME is set in the environment
        if not os.environ.get("ENV_NAME"):
            raise Exception("ENV_NAME is not set in the environment")


config = NetBoxPaas



import requests
import semver
import yaml
from django.conf import settings
from django.utils.safestring import mark_safe

from .constants import NETBOX_JPS_REPO

__all__ = ("get_plugins_list", "filter_releases", "job_msg", "is_upgrade_available")


def get_plugins_list():
    plugins = {}

    # Download plugins.yaml from GitHub
    response = requests.get(f"{NETBOX_JPS_REPO}/plugins.yaml")
    if response.ok:
        plugins = yaml.safe_load(response.text)

    return plugins


def is_compatible(netbox_version, min_version, max_version):
    """
    Check if the NetBox version is compatible with the plugin.
    """
    if min_version and semver.compare(netbox_version, min_version) < 0:
        return False

    if max_version and semver.compare(netbox_version, max_version) > 0:
        return False

    return True


def filter_releases(plugin, netbox_version: str = None):
    """
    Filter the releases based on the NetBox version.
    """
    compatible_releases = []
    version = netbox_version or settings.VERSION

    for release in plugin.get("releases", []):
        netbox = release.get("netbox")
        min_version = netbox.get("min")
        max_version = netbox.get("max")

        if is_compatible(version, min_version, max_version):
            compatible_releases.append(release["tag"])

    try:
        return sorted(compatible_releases, key=semver.parse_version_info, reverse=True)
    except ValueError:
        return compatible_releases


def is_upgrade_available(plugin, plugin_version: str):
    """
    Check if a newer version of the plugin is available.
    """
    releases = filter_releases(plugin, settings.VERSION)
    if not releases:
        return False

    # Check if the latest release is newer than the current version
    latest_release = releases[0]
    latest_release = latest_release.lstrip("v")
    return semver.compare(latest_release, plugin_version) > 0


def job_msg(job):
    return mark_safe(f"Job <a href='{job.get_absolute_url()}'>{job}</a> has been created successfully.")



from drf_spectacular.utils import extend_schema_field
from drf_spectacular.types import OpenApiTypes

from django.conf import settings
from importlib.metadata import metadata
from rest_framework import serializers
from netbox.api.serializers import NetBoxModelSerializer, WritableNestedSerializer

from ..models import *
from ..utils import is_upgrade_available, filter_releases

__all__ = [
    "NetBoxConfigurationSerializer",
    "NetBoxDBBackupSerializer",
    "NestedNetBoxConfigurationSerializer",
    "NetBoxPluginSerializer",
    "NetBoxPluginInstallSerializer",
]


class NetBoxConfigurationSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="plugins-api:netbox_paas-api:netboxconfiguration-detail")

    env_name = serializers.CharField(
        read_only=True,
    )
    key = serializers.CharField(
        write_only=True,
    )
    license = serializers.CharField(write_only=True, allow_blank=True)

    class Meta:
        model = NetBoxConfiguration
        fields = (
            "id",
            "url",
            "display",
            "env_name",
            "env_name_storage",
            "license",
            "key",
            "description",
            "custom_fields",
            "comments",
            "created",
            "last_updated",
        )


class NestedNetBoxConfigurationSerializer(WritableNestedSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="plugins-api:netbox_paas-api:netboxconfiguration-detail")

    class Meta:
        model = NetBoxConfiguration
        fields = ("id", "url", "display", "env_name", "env_name_storage")


class NetBoxDBBackupSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="plugins-api:netbox_paas-api:netboxdbbackup-detail")

    netbox_env = NestedNetBoxConfigurationSerializer()

    class Meta:
        model = NetBoxDBBackup
        fields = ("id", "url", "display", "netbox_env", "crontab", "keep_backups")


class NetBoxPluginOwnerSerializer(serializers.Serializer):
    name = serializers.CharField()
    url = serializers.URLField()

    class Meta:
        fields = (
            "name",
            "url",
        )


class NetBoxPluginSerializer(serializers.Serializer):
    plugin_id = serializers.CharField()
    name = serializers.CharField()
    display = serializers.SerializerMethodField(read_only=True)
    app_label = serializers.CharField()
    description = serializers.CharField()
    github_api_url = serializers.URLField()
    github_url = serializers.URLField()
    owner = NetBoxPluginOwnerSerializer()
    private = serializers.BooleanField()
    pypi_url = serializers.URLField(
        required=False,
    )
    releases = serializers.SerializerMethodField(
        required=False,
    )
    is_installed = serializers.SerializerMethodField(
        read_only=True,
    )
    current_version = serializers.SerializerMethodField(
        read_only=True,
    )
    upgrade_available = serializers.SerializerMethodField(
        read_only=True,
    )

    @extend_schema_field(OpenApiTypes.STR)
    def get_display(self, obj):
        return obj["name"]

    @extend_schema_field(OpenApiTypes.BOOL)
    def get_is_installed(self, obj):
        return obj["app_label"] in settings.PLUGINS

    @extend_schema_field(OpenApiTypes.STR)
    def get_current_version(self, obj):
        if not self.get_is_installed(obj):
            return None

        return metadata(obj["app_label"]).get("Version", None)

    @extend_schema_field(OpenApiTypes.BOOL)
    def get_upgrade_available(self, obj):
        if not self.get_is_installed(obj):
            return False

        plugin = self.context['plugins'][obj["plugin_id"]]
        return is_upgrade_available(plugin, self.get_current_version(obj))

    def get_releases(self, obj):
        plugin = self.context['plugins'][obj["plugin_id"]]
        return filter_releases(plugin)


class NetBoxPluginInstallSerializer(serializers.Serializer):
    plugin_id = serializers.CharField()
    version = serializers.CharField()
    configuration = serializers.JSONField(
        required=False,
    )

    class Meta:
        fields = (
            "plugin_id",
            "version",
            "configuration",
        )

    def validate_plugin_id(self, value):
        if value not in self.context['plugins']:
            raise serializers.ValidationError(f"Plugin '{value}' not found.")

        return value



import random
import string
from importlib.metadata import metadata

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.shortcuts import redirect, get_object_or_404, render
from django.urls import reverse
from django.views.generic import View

from netbox.views import generic
from utilities.utils import normalize_querydict
from utilities.views import register_model_view, GetReturnURLMixin
from . import constants, forms, models, tables, utils


@register_model_view(models.NetBoxConfiguration)
class NetBoxConfigurationView(generic.ObjectView):
    queryset = models.NetBoxConfiguration.objects.all()


class NetBoxConfigurationListView(generic.ObjectListView):
    queryset = models.NetBoxConfiguration.objects.all()

    def get(self, request):
        if obj := models.NetBoxConfiguration.objects.first():
            return redirect("plugins:netbox_paas:netboxconfiguration", pk=obj.pk)

        return redirect("plugins:netbox_paas:netboxconfiguration_add")


@register_model_view(models.NetBoxConfiguration, "edit")
class NetBoxConfigurationEditView(generic.ObjectEditView):
    queryset = models.NetBoxConfiguration.objects.all()
    form = forms.NetBoxConfigurationForm


@register_model_view(models.NetBoxConfiguration, "logs")
class NetBoxNodeLog(PermissionRequiredMixin, View):
    def get_permission_required(self):
        return ["netbox_paas.view_netboxconfiguration"]

    def post(self, request, pk):
        """
        Retrieve logs from Jelastic and return them as a response.
        """
        instance = get_object_or_404(models.NetBoxConfiguration, pk=pk)
        env_name = request.POST.get("env_name")
        node_id = request.POST.get("node_id")

        paas = instance.paas(env_name)

        node = paas.get_node(node_id)
        logs = paas.get_node_log(node_id)
        return render(
            request,
            "netbox_paas/nodelogs.html",
            {
                "object": instance,
                "env_name": env_name,
                "node": node,
                "logs": logs,
            },
        )


@register_model_view(models.NetBoxConfiguration, "settings")
class NetBoxSettingsView(PermissionRequiredMixin, GetReturnURLMixin, View):
    def get_permission_required(self):
        return ["netbox_paas.change_netboxconfiguration"]

    def get(self, request, *args, **kwargs):
        obj = get_object_or_404(models.NetBoxConfiguration, pk=kwargs["pk"])
        form = forms.NetBoxSettingsForm(initial=obj.netbox_settings())

        return render(
            request,
            "generic/object_edit.html",
            {
                "model": models.NetBoxConfiguration,
                "object": obj,
                "form": form,
                "return_url": reverse(
                    "plugins:netbox_paas:netboxconfiguration",
                    kwargs={"pk": obj.pk},
                ),
            },
        )

    def post(self, request, *args, **kwargs):
        obj = get_object_or_404(models.NetBoxConfiguration, pk=kwargs["pk"])

        form = forms.NetBoxSettingsForm(request.POST, initial=obj.netbox_settings())
        if form.is_valid():
            # Call Jelastic to apply the ENV_VARs on all NetBox nodes (including workers)
            job = obj.enqueue(
                obj.apply_settings,
                request,
                data=form.cleaned_data,
            )
            messages.success(request, utils.job_msg(job))
            return redirect("core:job", pk=job.pk)

        return render(
            request,
            "generic/object_edit.html",
            {
                "object": obj,
                "form": form,
                "return_url": self.get_return_url(request, obj),
            },
        )


@register_model_view(models.NetBoxConfiguration, "restart")
class NetBoxRestartView(PermissionRequiredMixin, View):
    def get_permission_required(self):
        return ["netbox_paas.change_netboxconfiguration"]

    def post(self, request, pk):
        """
        Restarts container by node group.
        """
        instance = get_object_or_404(models.NetBoxConfiguration, pk=pk)
        env_name = request.POST.get("env_name")
        node_group = request.POST.get("node_group")
        job = instance.enqueue(
            instance.paas(env_name, auto_init=False).restart_nodes,
            request,
            node_groups=[node_group],
        )
        messages.success(request, utils.job_msg(job))
        return redirect("plugins:netbox_paas:netboxconfiguration", pk=instance.pk)


@register_model_view(models.NetBoxConfiguration, "backup_storage", path="backup-storage")
class NetBoxStorageView(PermissionRequiredMixin, GetReturnURLMixin, View):
    def get_permission_required(self):
        return ["netbox_paas.view_netboxconfiguration"]

    def get(self, request, *args, **kwargs):
        obj = get_object_or_404(models.NetBoxConfiguration, pk=kwargs["pk"])
        form = forms.NetBoxBackupStorageForm()

        return render(
            request,
            "generic/object_edit.html",
            {
                "model": models.NetBoxConfiguration,
                "object": obj,
                "form": form,
                "return_url": reverse(
                    "plugins:netbox_paas:netboxconfiguration",
                    kwargs={"pk": obj.pk},
                ),
            },
        )

    def post(self, request, *args, **kwargs):
        obj = get_object_or_404(models.NetBoxConfiguration, pk=kwargs["pk"])

        form = forms.NetBoxBackupStorageForm(request.POST)
        if form.is_valid():
            # Generate a random env name
            env_name = "env-"
            env_name += "".join([random.choice(string.digits) for _ in range(7)])

            # Deploy a new environment for backup-storage
            job = obj.enqueue(
                obj.paas(env_name, auto_init=False).client.marketplace.App.Install,
                request,
                id="wp-restore",
                env_name=env_name,
                settings={
                    "customName": form.cleaned_data["deployment"],
                    "storageNodesCount": form.cleaned_data["node_count"],
                    "diskspace": form.cleaned_data["storage_size"],
                },
                display_name=form.cleaned_data["display_name"],
                region=form.cleaned_data["region"],
                skip_email=True,
            )

            obj.env_name_storage = env_name
            obj.save()

            messages.success(request, utils.job_msg(job))
            return redirect("core:job", pk=job.pk)

        return render(
            request,
            "generic/object_edit.html",
            {
                "object": obj,
                "form": form,
                "return_url": self.get_return_url(request, obj),
            },
        )


@register_model_view(models.NetBoxDBBackup)
class NetBoxDBBackupView(generic.ObjectView):
    queryset = models.NetBoxDBBackup.objects.all()

    def get_extra_context(self, request, instance):
        table = tables.NetBoxBackupsTable(instance.list_backups())

        return {"backup_table": table}


class NetBoxDBBackupListView(generic.ObjectListView):
    queryset = models.NetBoxDBBackup.objects.all()

    def get(self, request):
        if obj := models.NetBoxDBBackup.objects.first():
            return redirect("plugins:netbox_paas:netboxdbbackup", pk=obj.pk)

        return redirect("plugins:netbox_paas:netboxdbbackup_add")


@register_model_view(models.NetBoxDBBackup, "edit")
class NetBoxDBBackupEditView(generic.ObjectEditView):
    queryset = models.NetBoxDBBackup.objects.all()
    form = forms.NetBoxDBBackupForm


@register_model_view(models.NetBoxDBBackup, "delete")
class NetBoxDBBackupDeleteView(generic.ObjectDeleteView):
    queryset = models.NetBoxDBBackup.objects.all()


@register_model_view(models.NetBoxDBBackup, "backup")
class NetBoxDBBackupBackupView(PermissionRequiredMixin, View):
    def get_permission_required(self):
        return ["netbox_paas.change_netboxdbbackup"]

    def post(self, request, pk):
        """
        Create a new backup.
        """
        instance = get_object_or_404(models.NetBoxDBBackup, pk=pk)

        job = instance.backup(request)
        messages.success(request, utils.job_msg(job))
        return redirect("core:job", pk=job.pk)


@register_model_view(models.NetBoxDBBackup, "restore")
class NetBoxDBBackupRestoreView(PermissionRequiredMixin, View):
    def get_permission_required(self):
        return ["netbox_paas.change_netboxdbbackup"]

    def post(self, request, pk):
        """
        Restore a backup.
        """
        instance = get_object_or_404(models.NetBoxDBBackup, pk=pk)
        backup_name = request.POST.get("name")

        job = instance.restore(request, backup_name)
        messages.success(request, utils.job_msg(job))
        return redirect("plugins:netbox_paas:netboxdbbackup", pk=instance.pk)


class NetBoxPluginListView(View):
    def get(self, request):
        if nc := models.NetBoxConfiguration.objects.first():
            installed_plugins = settings.PLUGINS
            installed_plugins = [metadata(plugin).get("Name") for plugin in installed_plugins]

            # Get the list of disabled plugins from NetBox
            disabled_plugins = [
                metadata(plugin).get("Name")
                for plugin in list(nc.get_env().load_plugins(file_name=constants.DISABLED_PLUGINS_FILE_NAME).keys())
            ]

            plugins = utils.get_plugins_list()
            for plugin_name, _ in plugins.items():
                if plugin_name in installed_plugins:
                    plugin_version = metadata(plugin_name).get("Version")
                    if upgrade_available := utils.is_upgrade_available(plugins[plugin_name], plugin_version):
                        plugins[plugin_name].update(
                            {
                                "upgrade_available": upgrade_available,
                                "latest_version": utils.filter_releases(plugins[plugin_name])[0],
                            }
                        )

                    plugins[plugin_name].update(
                        {
                            "installed": True,
                            "current_version": plugin_version,
                        }
                    )
                if plugin_name in disabled_plugins:
                    plugins[plugin_name].update(
                        {"installed": True, "disabled": True, "current_version": metadata(plugin_name).get("Version")}
                    )

            # Divide the plugins into two lists: installed and not installed
            plugins = {
                "installed": {
                    plugin_name: plugin for plugin_name, plugin in plugins.items() if plugin.get("installed")
                },
                "not_installed": {
                    plugin_name: plugin for plugin_name, plugin in plugins.items() if not plugin.get("installed")
                },
            }

            # Divide not installed plugins into two lists: subscription and community
            plugins["not_installed"] = {
                "subscription": {
                    plugin_name: plugin
                    for plugin_name, plugin in plugins["not_installed"].items()
                    if plugin.get("private")
                },
                "community": {
                    plugin_name: plugin
                    for plugin_name, plugin in plugins["not_installed"].items()
                    if not plugin.get("private")
                },
            }

            return render(
                request,
                "netbox_paas/plugins_store.html",
                {"object": nc, "plugins": plugins},
            )

        messages.error(request, "You must configure NetBox first.")
        return redirect("plugins:netbox_paas:netboxconfiguration_add")


@register_model_view(models.NetBoxConfiguration, "upgrades")
class NetBoxPluginUpgradesView(PermissionRequiredMixin, GetReturnURLMixin, View):
    queryset = models.NetBoxConfiguration.objects.all()
    form = forms.NetBoxUpgradeForm

    def get_permission_required(self):
        return ["netbox_paas.change_netboxconfiguration"]

    def get(self, request, pk, *args, **kwargs):
        obj = get_object_or_404(models.NetBoxConfiguration, pk=pk)

        return render(
            request,
            "generic/object_edit.html",
            {
                "object": obj,
                "form": self.form(),
            },
        )

    def post(self, request, pk, *args, **kwargs):
        obj = get_object_or_404(models.NetBoxConfiguration, pk=pk)
        form = self.form(request.POST)

        if form.is_valid():
            job = obj.enqueue(
                obj.get_env().upgrade,
                request,
                version=form.cleaned_data["version"],
                lic=obj.license,
            )
            messages.success(request, utils.job_msg(job))
            return redirect("core:job", pk=job.pk)

        return render(
            request,
            "generic/object_edit.html",
            {
                "object": obj,
                "form": form,
            },
        )


@register_model_view(models.NetBoxConfiguration, "plugin_install", path="plugin-install")
class NetBoxPluginInstallView(generic.ObjectEditView):
    queryset = models.NetBoxConfiguration.objects.all()
    form = forms.NetBoxPluginInstallForm

    def initialize_form(self, request):
        data = request.POST if request.method == "POST" else None
        initial_data = normalize_querydict(request.GET)

        form = self.form(data=data, initial=initial_data)

        return form

    def get(self, request, *args, **kwargs):
        obj = self.get_object(**kwargs)
        obj = self.alter_object(obj, request, args, kwargs)
        model = self.queryset.model

        form = self.initialize_form(request)
        return render(
            request,
            self.template_name,
            {
                "model": model,
                "object": obj,
                "form": form,
                "return_url": self.get_return_url(request, obj),
                **self.get_extra_context(request, obj),
            },
        )

    def post(self, request, *args, **kwargs):
        obj = self.get_object(**kwargs)
        form = self.initialize_form(request)

        if form.is_valid():
            plugin = utils.get_plugins_list().get(form.cleaned_data["name"])

            job = obj.enqueue(
                obj.get_env().install_plugin,
                request,
                plugin=plugin,
                version=form.cleaned_data["version"],
                plugin_settings=form.cleaned_data["configuration"],
                github_token=obj.license,
            )
            messages.success(request, utils.job_msg(job))
            return redirect("core:job", pk=job.pk)

        return render(
            request,
            self.template_name,
            {
                "object": obj,
                "form": form,
                "return_url": self.get_return_url(request, obj),
                **self.get_extra_context(request, obj),
            },
        )


@register_model_view(models.NetBoxConfiguration, "plugin_uninstall", path="plugin-uninstall")
class NetBoxPluginUninstallView(generic.ObjectDeleteView):
    queryset = models.NetBoxConfiguration.objects.all()
    template_name = "netbox_paas/plugin_uninstall.html"

    def get(self, request, *args, **kwargs):
        obj = self.get_object(**kwargs)
        plugin = utils.get_plugins_list().get(request.GET.get("name"))

        if plugin is None:
            messages.error(request, "Plugin not found.")
            return redirect("plugins:netbox_paas:netboxplugin_list")

        form = forms.ConfirmationForm(initial=request.GET)

        return render(
            request,
            self.template_name,
            {
                "object": obj,
                "form": form,
                "return_url": self.get_return_url(request, obj),
                **self.get_extra_context(request, obj),
            },
        )

    def post(self, request, *args, **kwargs):
        obj = self.get_object(**kwargs)
        form = forms.ConfirmationForm(request.POST)

        if form.is_valid():
            plugin = utils.get_plugins_list().get(form.cleaned_data["name"])
            job = obj.enqueue(obj.get_env().uninstall_plugin, request, plugin=plugin)

            messages.success(request, utils.job_msg(job))
            return redirect("core:job", pk=job.pk)

        return render(
            request,
            self.template_name,
            {
                "object": obj,
                "form": form,
                "return_url": self.get_return_url(request, obj),
                **self.get_extra_context(request, obj),
            },
        )


@register_model_view(models.NetBoxConfiguration, "plugin_enable", path="plugin-enable")
class NetBoxPluginEnableView(generic.ObjectDeleteView):
    queryset = models.NetBoxConfiguration.objects.all()
    template_name = "netbox_paas/plugin_enable.html"

    def get(self, request, *args, **kwargs):
        obj = self.get_object(**kwargs)
        plugin = utils.get_plugins_list().get(request.GET.get("name"))

        if plugin is None:
            messages.error(request, "Plugin not found.")
            return redirect("plugins:netbox_paas:netboxplugin_list")

        form = forms.ConfirmationForm(initial=request.GET)

        return render(
            request,
            self.template_name,
            {
                "object": obj,
                "form": form,
                "return_url": self.get_return_url(request, obj),
                **self.get_extra_context(request, obj),
            },
        )

    def post(self, request, *args, **kwargs):
        obj = self.get_object(**kwargs)
        form = forms.ConfirmationForm(request.POST)

        if form.is_valid():
            plugin = utils.get_plugins_list().get(form.cleaned_data["name"])
            job = obj.enqueue(obj.get_env().enable_plugin, request, plugin=plugin)

            messages.success(request, utils.job_msg(job))
            return redirect("core:job", pk=job.pk)

        return render(
            request,
            self.template_name,
            {
                "object": obj,
                "form": form,
                "return_url": self.get_return_url(request, obj),
                **self.get_extra_context(request, obj),
            },
        )


@register_model_view(models.NetBoxConfiguration, "plugin_disable", path="plugin-disable")
class NetBoxPluginDisableView(generic.ObjectDeleteView):
    queryset = models.NetBoxConfiguration.objects.all()
    template_name = "netbox_paas/plugin_disable.html"

    def get(self, request, *args, **kwargs):
        obj = self.get_object(**kwargs)
        plugin = utils.get_plugins_list().get(request.GET.get("name"))

        if plugin is None:
            messages.error(request, "Plugin not found.")
            return redirect("plugins:netbox_paas:netboxplugin_list")

        form = forms.ConfirmationForm(initial=request.GET)

        return render(
            request,
            self.template_name,
            {
                "object": obj,
                "form": form,
                "return_url": self.get_return_url(request, obj),
                **self.get_extra_context(request, obj),
            },
        )

    def post(self, request, *args, **kwargs):
        obj = self.get_object(**kwargs)
        form = forms.ConfirmationForm(request.POST)

        if form.is_valid():
            plugin = utils.get_plugins_list().get(form.cleaned_data["name"])
            job = obj.enqueue(obj.get_env().disable_plugin, request, plugin=plugin)

            messages.success(request, utils.job_msg(job))
            return redirect("core:job", pk=job.pk)

        return render(
            request,
            self.template_name,
            {
                "object": obj,
                "form": form,
                "return_url": self.get_return_url(request, obj),
                **self.get_extra_context(request, obj),
            },
        )



from extras.plugins import PluginMenu
from extras.plugins import PluginMenuItem

menu = PluginMenu(
    label="Netbox PAAS",
    icon_class="mdi mdi-cloud",
    groups=(
        (
            "",
            (
                PluginMenuItem(
                    link_text="Manage",
                    link="plugins:netbox_paas:netboxconfiguration_list",
                    permissions=["netbox_paas.view_netboxconfiguration"],
                ),
                PluginMenuItem(
                    link_text="DB Backups",
                    link="plugins:netbox_paas:netboxdbbackup_list",
                    permissions=["netbox_paas.change_netboxconfiguration"],
                ),
                PluginMenuItem(
                    link_text="Plugins Store",
                    link="plugins:netbox_paas:netboxplugin_list",
                    permissions=["netbox_paas.view_netboxconfiguration"],
                ),
            ),
        ),
    ),
)



# Generated by Django 4.2.8 on 2024-05-10 05:43

from django.db import migrations, models
import django.db.models.deletion


def migrate_models(apps, schema_editor):
    try:
        OldModel1 = apps.get_model('netbox_cloud_pilot', 'NetBoxConfiguration')
        OldModel2 = apps.get_model('netbox_cloud_pilot', 'NetBoxDBBackup')
    except LookupError:
        # Skip if the old models don't exist
        return

    NewModel1 = apps.get_model('netbox_paas', 'NetBoxConfiguration')
    NewModel2 = apps.get_model('netbox_paas', 'NetBoxDBBackup')

    # Migrate data from old models to new models
    for old_instance in OldModel1.objects.all():
        new_instance = NewModel1(
            id=old_instance.id,
            key=old_instance.key,
            env_name_storage=old_instance.env_name_storage,
            license=old_instance.license,
            description=old_instance.description,
            created=old_instance.created,
            last_updated=old_instance.last_updated,
            custom_field_data=old_instance.custom_field_data,
            comments=old_instance.comments,
        )
        new_instance.save()

    for old_instance in OldModel2.objects.all():
        new_instance = NewModel2(
            id=old_instance.id,
            created=old_instance.created,
            last_updated=old_instance.last_updated,
            crontab=old_instance.crontab,
            keep_backups=old_instance.keep_backups,
            netbox_env_id=old_instance.netbox_env.id,
        )
        new_instance.save()


class Migration(migrations.Migration):
    dependencies = [
        ('netbox_paas', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(migrate_models),
    ]



from django.urls import include, path

from utilities.urls import get_model_urls
from . import views

app_name = "netbox_paas"

urlpatterns = (
    # Configurations
    path(
        "configurations/",
        views.NetBoxConfigurationListView.as_view(),
        name="netboxconfiguration_list",
    ),
    path(
        "configurations/add/",
        views.NetBoxConfigurationEditView.as_view(),
        name="netboxconfiguration_add",
    ),
    path(
        "configurations/<int:pk>/",
        include(get_model_urls(app_name, "netboxconfiguration")),
    ),
    # DB Backups
    path(
        "db-backups/",
        views.NetBoxDBBackupListView.as_view(),
        name="netboxdbbackup_list",
    ),
    path(
        "db-backups/add/",
        views.NetBoxDBBackupEditView.as_view(),
        name="netboxdbbackup_add",
    ),
    path(
        "db-backups/<int:pk>/",
        include(get_model_urls(app_name, "netboxdbbackup")),
    ),
    # Plugins
    path(
        "plugins/",
        views.NetBoxPluginListView.as_view(),
        name="netboxplugin_list",
    ),
)



from django import forms

__all__ = (
    "NbSettings",
    "NbSection",
    "Param",
)


class Base:
    def __iter__(self):
        for key in self.__dict__:
            yield key, getattr(self, key)

    def __getitem__(self, item):
        return getattr(self, item)

    def __str__(self):
        return f"{self.__class__.__name__} {self.__dict__}"

    def __repr__(self):
        class_name = self.__class__.__name__
        repr_args = ", ".join([f"{k}={repr(v)}" for k, v in self])
        return f"{class_name}({repr_args})"

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()


class Param(Base):
    def __init__(
        self,
        key: str,
        label: str,
        field=None,
        field_kwargs: dict = None,
        help_text: str = None,
        placeholder: str = None,
        required: bool = True,
        initial=None,
        choices=None,
    ):
        self.key = key
        self.label = label
        self.help_text = help_text
        self.required = required
        self.field = field or forms.CharField
        self.initial = initial
        self.choices = choices

        # Default field_kwargs if none provided
        default_kwargs = {"widget": forms.TextInput(attrs={"class": "form-control"})}

        # Update or set field_kwargs
        if field_kwargs is not None:
            widget = field_kwargs.get("widget")
            if widget and hasattr(widget, "attrs"):
                # Update the class attribute, retaining existing classes
                existing_classes = widget.attrs.get("class", "")
                widget.attrs["class"] = " ".join(filter(None, [existing_classes, "form-control"]))
            else:
                # If widget is not provided or doesn't have attrs, use the default
                field_kwargs.update(default_kwargs)
        else:
            field_kwargs = default_kwargs

        if placeholder:
            field_kwargs["widget"].attrs["placeholder"] = placeholder

        self.field_kwargs = field_kwargs


class NbSection(Base):
    def __init__(self, name: str | None, params: list[Param]):
        self.name = name
        self.params = params


class NbSettings(Base):
    def __init__(self, sections: list[NbSection]):
        self.sections = sections



from django.conf import settings
from drf_spectacular.utils import extend_schema
from drf_spectacular.openapi import OpenApiParameter, OpenApiTypes
from rest_framework.viewsets import ViewSet
from rest_framework.response import Response
from rest_framework.exceptions import ValidationError

from core.api.serializers import JobSerializer
from netbox.api.viewsets import NetBoxModelViewSet
from utilities.rqworker import get_workers_for_queue
from .serializers import *
from ..models import *
from ..utils import get_plugins_list


class NetBoxConfigurationViewSet(NetBoxModelViewSet):
    queryset = NetBoxConfiguration.objects.all()
    serializer_class = NetBoxConfigurationSerializer


class NetBoxDBBackupViewSet(NetBoxModelViewSet):
    queryset = NetBoxDBBackup.objects.all()
    serializer_class = NetBoxDBBackupSerializer


class NetBoxPluginViewSet(ViewSet):
    queryset = NetBoxConfiguration.objects.all()

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        self.netbox_configuration = NetBoxConfiguration.objects.first()
        if not self.netbox_configuration:
            raise ValidationError("NetBoxConfiguration object not found.")

        # Get plugins list
        self.plugins = get_plugins_list()

    def check_actions(self):
        # Check if RQ worker is running
        if not get_workers_for_queue('default'):
            raise ValidationError(
                "No RQ workers operating on the 'default' queue are currently active in the environment."
            )

        # Check if there are active actions in the environment
        if self.netbox_configuration.get_env().get_actions():
            raise ValidationError("There are currently actions running on the environment.")

    @extend_schema(
        responses={200: NetBoxPluginSerializer(many=True)},
    )
    def list(self, request):
        # Get plugins list
        plugins_list = []

        for plugin_id, plugin in self.plugins.items():
            plugins_list.append({**plugin, 'plugin_id': plugin_id})

        data = NetBoxPluginSerializer(
            plugins_list, many=True, context={'plugins': self.plugins, 'request': request}
        ).data
        return Response({'results': data, 'count': len(data)})

    @extend_schema(
        responses={200: NetBoxPluginSerializer},
        description="Get a plugin object.",
        parameters=[
            OpenApiParameter(
                name='id',
                type=OpenApiTypes.STR,
                location=OpenApiParameter.PATH,
                description="Plugin ID",
            ),
        ],
    )
    def retrieve(self, request, pk=None):
        # Get plugins list
        plugin = self.plugins.get(pk)
        if not plugin:
            raise ValidationError("Plugin not found.")

        data = NetBoxPluginSerializer(
            {
                **plugin,
                'plugin_id': pk,
            },
            context={'plugins': self.plugins, 'request': request},
        ).data
        return Response(data)

    @extend_schema(
        request=NetBoxPluginInstallSerializer,
        responses={200: JobSerializer},
        description="Update a plugin.",
        parameters=[
            OpenApiParameter(
                name='id',
                type=OpenApiTypes.STR,
                location=OpenApiParameter.PATH,
                description="Plugin ID",
            ),
        ],
    )
    def update(self, request, pk=None):
        plugin = self.plugins.get(pk)
        serializer = NetBoxPluginInstallSerializer(
            data=request.data, context={'plugins': self.plugins, 'request': request}
        )
        serializer.is_valid(raise_exception=True)

        self.check_actions()

        job = self.netbox_configuration.enqueue(
            self.netbox_configuration.get_env().install_plugin,
            request,
            plugin=plugin,
            version=serializer.validated_data["version"],
            plugin_settings=serializer.validated_data.get("configuration"),
            github_token=self.netbox_configuration.license,
        )

        return Response(JobSerializer(job, context={'request': request}).data)

    @extend_schema(
        responses={200: JobSerializer},
        description="Uninstall a plugin.",
        parameters=[
            OpenApiParameter(
                name='id',
                type=OpenApiTypes.STR,
                location=OpenApiParameter.PATH,
                description="Plugin ID",
            ),
        ],
    )
    def destroy(self, request, pk=None):
        plugin = self.plugins.get(pk)
        if not plugin:
            raise ValidationError("Plugin not found.")

        if plugin['app_label'] not in settings.PLUGINS:
            raise ValidationError(f"Plugin {pk} is not installed.")

        self.check_actions()

        job = self.netbox_configuration.enqueue(
            self.netbox_configuration.get_env().uninstall_plugin,
            request,
            plugin=plugin,
        )

        return Response(JobSerializer(job, context={'request': request}).data)

    @extend_schema(
        request=NetBoxPluginInstallSerializer,
        responses={200: JobSerializer},
    )
    def create(self, request):
        serializer = NetBoxPluginInstallSerializer(
            data=request.data, context={'plugins': self.plugins, 'request': request}
        )
        serializer.is_valid(raise_exception=True)

        self.check_actions()

        plugin = self.plugins.get(serializer.validated_data["plugin_id"])
        job = self.netbox_configuration.enqueue(
            self.netbox_configuration.get_env().install_plugin,
            request,
            plugin=plugin,
            version=serializer.validated_data["version"],
            plugin_settings=serializer.validated_data.get("configuration"),
            github_token=self.netbox_configuration.license,
        )
        return Response(JobSerializer(job, context={'request': request}).data)



import ast
import json
import logging
import os
import re
from datetime import datetime
from functools import lru_cache

from croniter import croniter
from croniter.croniter import CroniterError
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.core.validators import MinLengthValidator, MaxLengthValidator
from django.db import models
from django.urls import reverse
from jelastic.api.exceptions import JelasticApiError
from netbox.models import ChangeLoggedModel, PrimaryModel
from netbox.models.features import JobsMixin

from .constants import (
    NETBOX_SUPERUSER_SETTINGS,
    NETBOX_SETTINGS,
    NODE_GROUP_SQLDB,
    NODE_GROUP_CP,
)
from .paas import *

__all__ = ("NetBoxConfiguration", "NetBoxDBBackup")

logger = logging.getLogger(__name__)


class NetBoxConfiguration(JobsMixin, PrimaryModel):
    """
    NetBoxConfig is a model that represents the configuration of NetBox.
    """

    key = models.CharField(max_length=255, unique=True, validators=[MinLengthValidator(40)])

    env_name_storage = models.CharField(
        max_length=255,
        unique=True,
        blank=True,
        validators=[MinLengthValidator(5)],
        verbose_name="Environment Name Storage",
    )

    license = models.CharField(
        max_length=255,
        validators=[MinLengthValidator(40), MaxLengthValidator(93)],
        blank=True,
        null=True,
    )

    class Meta:
        verbose_name = "NetBox Configuration"
        verbose_name_plural = "NetBox Configurations"

    def __str__(self):
        return self.env_name

    @property
    def env_name(self):
        return os.environ.get("ENV_NAME")

    def get_absolute_url(self):
        return reverse("plugins:netbox_paas:netboxconfiguration", args=[self.pk])

    def clean(self):
        if self.__class__.objects.exists() and not self.pk:
            raise ValidationError("There can only be one NetBoxConfiguration instance.")

        # Clear any cached values from lru_cache
        self.paas.cache_clear()

        # Environment name must not contain dots
        if "." in self.env_name:
            raise ValidationError("Environment name must not contain dots.")

        if self.pk:
            try:
                # Ensure the provided env_name exists
                self.paas(self.env_name)
            except JelasticApiError as e:
                raise ValidationError(e)

        if self.env_name_storage:
            try:
                # Ensure the provided env_name exists
                self.paas(self.env_name_storage)
            except JelasticApiError as e:
                raise ValidationError(e)

        if self.license:
            if not self.license.startswith("ghp_") and not self.license.startswith("github_pat_"):
                raise ValidationError({"license": "Invalid license key"})

    @lru_cache(maxsize=2)
    def paas(self, env_name, auto_init=True):
        return PaaSNetBox(
            token=self.key,
            env_name=env_name,
            auto_init=auto_init,
        )

    def get_env(self):
        return self.paas(self.env_name)

    def get_docker_tag(self):
        return self.get_env().get_master_node(NODE_GROUP_CP).get("customitem", {}).get("dockerTag", "")

    def get_env_storage(self):
        return self.paas(self.env_name_storage)

    def netbox_admin(self):
        """
        Returns NetBox administration configuration.
        """
        paas = self.get_env()

        result = {"url": paas.get_url()}
        for setting in NETBOX_SUPERUSER_SETTINGS:
            result[setting] = paas.get_env_var(setting)

        return result

    def netbox_settings(self):
        """
        Returns NetBox settings.
        """
        config = {}
        env = self.get_env()

        # Iterate through each section in NETBOX_SETTINGS
        for section in NETBOX_SETTINGS.sections:
            section_name = section.name
            config[section_name] = []

            # Iterate through each setting in the section
            for param in section.params:
                initial = env.get_env_var(param.key, param.initial)

                # Alter list of strings to comma-separated string
                if isinstance(initial, (tuple, list)) and all([isinstance(x, str) for x in initial]):
                    initial = ", ".join(initial)

                param.initial = initial

                config[section_name].append(param)

        return config

    def apply_settings(self, data: dict):
        """
        Apply NetBox settings.
        """
        logger.debug(f"Applying NetBox settings: {data}")

        all_keys = [param.key for section in NETBOX_SETTINGS.sections for param in section.params]

        env = self.get_env()
        # Get all NetBox node groups
        node_groups = env.get_nb_node_groups()

        for node_group in node_groups:
            logger.debug(f"Applying settings to node group: {node_group}")

            node_group_name = node_group.get("name")
            env_vars = env.get_env_vars(node_group_name)

            # Filter env_vars with keys that exist in all_keys only
            env_vars = {k: v for k, v in env_vars.items() if k in all_keys}

            # Remove all filtered env_vars
            env.remove_env_vars(node_group_name, list(env_vars.keys()))

            for section, params in data.items():
                if section is None:
                    vars = {}
                    # These values are to be updated in the environment variables
                    for key, value in params.items():
                        if value:
                            vars[key] = value
                    env.add_env_vars(node_group_name, vars)

        # Update the files on the NODE_GROUP_CP with named sections
        file_content = ""
        for section, params in data.items():
            if section is not None:
                # Construct the file content
                for key, value in params.items():
                    if value:
                        try:
                            file_content += f"{key} = {ast.literal_eval(value)}\n"
                        except (ValueError, SyntaxError):
                            if isinstance(value, str):
                                file_content += f"{key} = '{value}'\n"
                            else:
                                file_content += f"{key} = {value}\n"

        file_path = f"/etc/netbox/config/extra.py"
        logger.debug(f"Writing file: {file_path}")
        # Create the section file
        env.client.environment.File.Write(
            env_name=self.env_name,
            path=file_path,
            body=file_content,
            node_group=NODE_GROUP_CP,
            master_only=True,
            is_append_mode=False,
        )

        # Schedule a restart of the NetBox environment
        self.schedule_restart()

    def enqueue(self, func, request, *args, **kwargs):
        return self.paas(self.env_name).enqueue(
            func,
            self,
            request,
            *args,
            **kwargs,
        )

    def schedule_restart(self):
        """
        Schedule a restart of the NetBox environment.
        """
        logger.debug(f"Scheduling restart of NetBox environment: {self.env_name}")
        env = self.get_env()
        env.restart_nodes(
            [group["name"] for group in env.get_nb_node_groups()],
            lazy=True,
        )


class NetBoxDBBackup(ChangeLoggedModel):
    netbox_env = models.ForeignKey(
        to=NetBoxConfiguration,
        on_delete=models.CASCADE,
        related_name="db_backups",
        verbose_name="NetBox Environment",
    )

    crontab = models.CharField(
        max_length=255,
        help_text=(
            "Crontab expression for the backup schedule."
            "See <a href='https://crontab.guru/' target='_blank'>crontab.guru</a> for more information."
        ),
        default="@daily",
    )

    keep_backups = models.PositiveIntegerField(
        default=1,
        help_text="Number of newest backups to keep during rotation.",
    )

    class Meta:
        verbose_name = "NetBox DB Backup"
        verbose_name_plural = "NetBox DB Backups"

    def get_absolute_url(self):
        return reverse("plugins:netbox_paas:netboxdbbackup", args=[self.pk])

    def __str__(self):
        return self.crontab

    def save(self, *args, **kwargs):
        self.netbox_env.enqueue(
            self.netbox_env.get_env().install_addon,
            None,
            app_id="db-backup",
            node_group=NODE_GROUP_SQLDB,
            addon_settings={
                "scheduleType": 3,
                "cronTime": self.cron,
                "storageName": self.netbox_env.env_name_storage,
                "backupCount": self.keep_backups,
                "dbuser": "webadmin",
                "dbpass": getattr(self, "_db_password", ""),
            },
        )

        super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        self.netbox_env.enqueue(
            self.netbox_env.get_env().uninstall_addon,
            None,
            app_id="db-backup",
            node_group=NODE_GROUP_SQLDB,
            search={"nodeType": "postgresql", "app_id": "db-backup"},
        )

        super().delete(*args, **kwargs)

    def clean(self):
        super().clean()

        if self.__class__.objects.exists() and not self.pk:
            raise ValidationError("There can only be one NetBoxDBBackup instance.")

        # Ensure netbox_env has a storage env connected before proceeding
        if not self.netbox_env.env_name_storage:
            raise ValidationError(
                {"netbox_env": "Add a backup storage environment to the NetBoxConfiguration instance."}
            )

        if self.keep_backups > 30:
            raise ValidationError({"keep_backups": "The maximum number of backups to keep is 30."})

        if self.keep_backups < 1:
            raise ValidationError({"keep_backups": "The minimum number of backups to keep is 1."})

        try:
            croniter(self.crontab)
        except CroniterError as e:
            raise ValidationError({"crontab": e})

    @property
    def cron(self):
        return " ".join(croniter(self.crontab).expressions)

    @property
    def get_master_node(self):
        node_groups = self.netbox_env.get_env_storage().get_node_groups()
        master_node = None
        for node_group in node_groups:
            master_node = node_group.get("node", {})
            break

        return master_node

    def list_backups(self):
        master_node = self.get_master_node
        if not master_node:
            return []

        # Always update restic
        self.netbox_env.get_env_storage().execute_cmd(
            node_id=master_node.get("id"), command="/usr/bin/restic self-update 2>&1"
        )

        result = self.netbox_env.get_env_storage().execute_cmd(
            node_id=master_node.get("id"), command="/root/getBackupsAllEnvs.sh"
        )[0]

        if backups := json.loads(result.get("out", "")).get("backups", {}).get(self.netbox_env.env_name, []):
            # Cache backups in case next time it returns empty
            cache.set(f"netbox_db_backups_{self.pk}", backups, timeout=60 * 60)
        else:
            # Return cached backups if any
            backups = cache.get(f"netbox_db_backups_{self.pk}", [])

        return sorted(
            [self.__parse_backup_name(backup) for backup in backups],
            key=lambda x: x.get("datetime"),
            reverse=True,
        )

    def __parse_backup_name(self, backup_name):
        """
        Parse the backup name and return a dictionary with the values.
        """
        # Initial split at the first underscore
        parts = backup_name.split("_", 1)
        # Parse the date from the first part
        date = datetime.strptime(parts[0], "%Y-%m-%d").date()

        # Parse the time from the second part
        parts = parts[1].split("-", 1)
        timestamp, timezone = parts[0].split("_")
        # Parse the time from the first part
        time = datetime.fromtimestamp(int(timestamp)).time()

        # Combine the date and time
        date = datetime.combine(date, time)

        # Parse the backup type and the database version
        if match := re.search(r"(\w+)\((.+)\)", parts[1]):
            backup_type = match.group(1)
            db_version = match.group(2).split("-", 1)[1]
        else:
            backup_type, db_version = None, None

        return {
            "name": backup_name,
            "datetime": date,
            "timezone": timezone,
            "backup_type": backup_type,
            "db_version": db_version,
        }

    def backup(self, request=None):
        """
        Create a new database backup.
        """
        # Get installed addon
        addon = self.netbox_env.get_env().get_installed_addon(app_id="db-backup", node_group=NODE_GROUP_SQLDB)

        return self.netbox_env.enqueue(
            self.netbox_env.get_env().db_backup,
            request,
            app_unique_name=addon.get("uniqueName"),
        )

    def restore(self, request, backup_name):
        """
        Restore a database backup.
        """
        # Get installed addon
        addon = self.netbox_env.get_env().get_installed_addon(app_id="db-backup", node_group=NODE_GROUP_SQLDB)

        return self.netbox_env.enqueue(
            self.netbox_env.get_env().execute_action,
            request,
            app_unique_name=addon.get("uniqueName"),
            action="restore",
            params={
                "backupDir": backup_name,
                "backupedEnvName": self.netbox_env.env_name,
            },
        )



import logging
import time
from functools import lru_cache
from typing import Tuple

import requests
import yaml
from django.conf import settings
from jelastic import Jelastic
from jelastic.api.exceptions import JelasticApiError
from requirements import parse as requirements_parse
from requirements.parser import Requirement
from semver.version import Version

from core.choices import JobStatusChoices
from core.models import Job
from netbox.config import get_config
from . import utils
from .constants import JELASTIC_API, NODE_GROUP_CP, NODE_GROUP_SQLDB, DISABLED_PLUGINS_FILE_NAME, PLUGINS_FILE_NAME

logger = logging.getLogger("netbox_paas")

__all__ = (
    "PaaS",
    "PaaSNetBox",
)


class RequirementsParser:
    """
    This class is used to parse the requirements.txt file.
    """

    def __init__(self, req_str: str):
        self.requirements = self.loads(req_str)

    @staticmethod
    def loads(req_str: str) -> list[Requirement]:
        """
        Parse the requirements.txt file.
        """
        return list(requirements_parse(req_str))

    def _exists(self, req: Requirement) -> bool:
        """
        Check if a requirement exists.
        """
        for r in self.requirements:
            if req.vcs:
                if r.vcs == req.vcs and r.uri == req.uri:
                    return True
            else:
                if r.name == req.name:
                    return True

        return False

    def add(self, req_str: str) -> bool:
        """
        Add a requirement to the requirements.txt file.
        """
        # Check if the requirement already exists
        req = self.loads(req_str)[0]

        # Update the requirement if it already exists, else add a new one
        for i, existing_req in enumerate(self.requirements):
            if existing_req.name == req.name:
                self.requirements[i] = req  # Update in-place
                return True

        self.requirements.append(req)  # Add new requirement
        return True

    def remove(self, req_str: str) -> bool:
        """
        Remove a requirement from the requirements.txt file.
        """
        # Check if the requirement exists
        req = self.loads(req_str)[0]

        # It is possible that req does not contain a version, so we need to check if it exists in the requirements
        if not self._exists(req):
            return False

        self.requirements = [r for r in self.requirements if r.name != req.name]
        return True

    def dumps(self) -> str:
        """
        Dump the requirements.txt file.
        """
        return "\n".join([str(r.line) for r in self.requirements])


class PaaSJob:
    @classmethod
    def enqueue(cls, func, model, request=None, *args, **kwargs):
        """
        Enqueue a job to be executed asynchronously.
        """
        logger.info(f"Enqueuing job {func.__name__}")

        kwargs["_func"] = func

        return Job.enqueue(
            cls._run_job,
            model,
            name=func.__name__,
            user=request.user if request else None,
            job_timeout=settings.RQ_DEFAULT_TIMEOUT,
            **kwargs,
        )

    @staticmethod
    def _run_job(job, *args, **kwargs):
        """
        Run a job.
        """
        data = {"params": kwargs}

        # Start the job
        try:
            job.start()
            func = kwargs.pop("_func")
            result = func(*args, **kwargs)

            data.update({"result": result})

            job.terminate()
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(e)
            data.update({"error": str(e)})
            job.terminate(status=JobStatusChoices.STATUS_ERRORED)

        job.data = data
        job.save()


class PaaS(PaaSJob):
    """
    This class is used to manage the PaaS layer of the Jelastic platform.
    """

    def __init__(self, token: str, env_name: str, auto_init: bool = True):
        self.admin_url = JELASTIC_API
        self.client = Jelastic(base_url=self.admin_url, token=token)
        self.env_name = env_name

        if auto_init:
            self._get_env_info()
        logger.debug(f"PaaS layer initialized for environment {self.env_name}")

    @lru_cache(maxsize=1)
    def _get_env_info(self):
        """
        Get information about the environment.
        """
        logger.debug("Getting environment information")
        return self.client.environment.Control.GetEnvInfo(env_name=self.env_name)

    @lru_cache(maxsize=10)
    def _get_env_var(self, node_group):
        """
        Get environment variables for a node group.
        """
        logger.debug(f"Getting environment variables for node group {node_group}")
        return self.client.environment.Control.GetContainerEnvVarsByGroup(
            env_name=self.env_name, node_group=node_group
        ).get("object", {})

    def clear_cache(self):
        """
        Clear the cache.
        """
        logger.debug("Clearing cache")
        self._get_env_info.cache_clear()
        self._get_env_var.cache_clear()

    def get_env(self):
        """
        Get the environment.
        """
        return self._get_env_info()["env"]

    def get_nodes(self, node_group: str = None, is_master: bool = True) -> dict | list:
        """
        Get the environment nodes.
        """
        logger.debug(f"Getting nodes for node group {node_group}, is_master={is_master}")
        nodes = self._get_env_info().get("nodes", [])
        if node_group:
            nodes = [node for node in nodes if node["nodeGroup"] == node_group]
        if is_master:
            try:
                nodes = [node for node in nodes if node["ismaster"]][0]
            except IndexError:
                nodes = {}
        return nodes

    def get_node_groups(self):
        """
        Get the environment node groups.
        """
        node_groups = sorted(
            self._get_env_info().get("nodeGroups", {}),
            key=lambda x: x.get("displayName", x.get("name")),
        )

        # For each node group, get the related nodes
        for node_group in node_groups:
            node_group["node"] = self.get_nodes(node_group=node_group["name"], is_master=True)

        return node_groups

    def get_node(self, node_id):
        """
        Get a node by its ID.
        """
        logger.debug(f"Getting node {node_id} for environment {self.env_name}")
        nodes = self._get_env_info().get("nodes", [])
        return next((node for node in nodes if str(node["id"]) == node_id), None)

    def get_external_domains(self):
        """
        Get the external domains.
        """
        logger.debug(f"Getting external domains for environment {self.env_name}")
        return self._get_env_info().get("extdomains", [])

    def is_ssl_enabled(self):
        """
        Check if SSL is enabled.
        """
        logger.debug(f"Checking if built-in SSL is enabled for environment {self.env_name}")
        return self.get_env().get("sslstate", False)

    def get_node_log(self, node_id, path="/var/log/run.log"):
        """
        Get the logs of a node.
        """
        logger.debug(f"Getting log for node {node_id} for environment {self.env_name}")
        return self.client.environment.Control.ReadLog(env_name=self.env_name, node_id=node_id, path=path).get(
            "body", ""
        )

    def get_url(self):
        """
        Get the environment URL.
        """
        logger.debug(f"Getting URL for environment {self.env_name}")
        scheme = "https" if self.is_ssl_enabled() else "http"
        if ext_domains := self.get_external_domains():
            return f"{scheme}://{ext_domains[0]}"

        domain = self.get_env().get("domain", "")
        return f"{scheme}://{domain}"

    def get_addons(self, node_group, search=None):
        """
        Get the addons for a node group.
        """
        logger.debug(f"Getting addons for node group {node_group}")
        return self.client.marketplace.App.GetAddonList(
            env_name=self.env_name, node_group=node_group, search=search
        ).get("apps", [])

    def get_installed_addon(self, app_id, node_group, search=None):
        """
        Get the installed addon for a node group.
        """
        logger.debug(f"Checking if addon ({app_id}) is installed for node group {node_group}")
        addons = self.get_addons(node_group=node_group, search=search)

        return next(
            (addon for addon in addons if addon["app_id"] == app_id and addon.get("isInstalled", False)),
            None,
        )

    def add_env_vars(self, node_group, env_vars):
        """
        Add environment variables to a node group.
        """
        logger.debug(f"Adding environment variables to node group {node_group}")
        return self.client.environment.Control.AddContainerEnvVars(
            env_name=self.env_name, node_group=node_group, vars=env_vars
        )

    def get_env_vars(self, node_group):
        """
        Get environment variables for a node group.
        """
        logger.debug(f"Getting environment variables for node group {node_group}")
        return self.client.environment.Control.GetContainerEnvVarsByGroup(
            env_name=self.env_name, node_group=node_group
        ).get("object", {})

    def remove_env_vars(self, node_group, env_vars: list[str]):
        """
        Remove environment variables from a node group.
        """
        logger.debug(f"Removing environment variables from node group {node_group}")
        return self.client.environment.Control.RemoveContainerEnvVars(
            env_name=self.env_name, node_group=node_group, vars=env_vars
        )

    def run_script(self, name, code, description=None, params=None, delay=10):
        """
        Run a script.
        """
        app_id = self.get_env().get("appid")
        delay = delay * 1000

        try:
            result = self.client.development.Scripting.GetScripts(
                app_id=app_id,
                type="js",
            )
            scripts = result.get("scripts", [])
            for script in scripts:
                if script.get("name") == name:
                    logger.debug(f"Deleting existing script {name}")
                    self.client.development.Scripting.DeleteScript(app_id=app_id, name=name)
                    continue
        except JelasticApiError as e:
            logger.error(e)

        self.client.development.Scripting.CreateScript(app_id=app_id, name=name, type="js", code=code)

        return self.client.utils.Scheduler.CreateEnvTask(
            env_name=self.env_name,
            script=name,
            trigger={"once_delay": delay},
            description=description,
            params=params,
        )

    def restart_nodes(self, node_groups: list[str], lazy: bool = False, delay: int = 10):
        """
        Restart nodes for a list of node groups.
        """
        # Sort the node groups in ascending order
        node_groups = sorted(node_groups)

        results = []

        logger.debug(f"Restarting nodes for node groups {', '.join(node_groups)}")
        if lazy:
            for node_group in node_groups:
                script_name = f"ncp-restart-{node_group}"
                script_code = """
                var c = jelastic.environment.control, e = envName, s = session, r, resp;
                resp = c.GetEnvInfo(e, s);
                if (resp.result != 0) return resp;
                r = c.RestartNodes({ envName: e, session: s, nodeGroup: nodeGroup, isSequential: true, manageDnsState: true, delay: 60000 });
                if (r.result != 0) return r;
                return { result: 0, message: 'Restarted ' + nodeGroup}
                """

                task_result = self.run_script(
                    name=script_name,
                    code=script_code,
                    description=f"Restart {node_group} nodes",
                    params={"envName": self.env_name, "nodeGroup": node_group},
                    delay=delay if node_group == NODE_GROUP_CP else delay * 10,
                )
                results.append(task_result)

            return results

        for node_group in node_groups:
            task_result = self.client.environment.Control.RestartNodes(
                env_name=self.env_name,
                node_group=node_group,
                is_sequential=True,
                manage_dns_state=True,
                delay=delay,
            )
            results.append(task_result)

        return results

    def execute_cmd(self, node_id, command):
        """
        Execute a command on a node.
        """
        logger.debug(f"Executing command {command} on node {node_id}")
        return self.client.environment.Control.ExecCmdById(
            env_name=self.env_name,
            node_id=node_id,
            command_list=[
                {
                    "command": command,
                }
            ],
        ).get("responses", [])

    def execute_action(self, app_unique_name, action="configure", params=None):
        """
        Execute an action on an addon.
        """
        logger.info(f"Executing action {action} on addon {app_unique_name}")
        return self.client.marketplace.Installation.ExecuteAction(
            app_unique_name=app_unique_name,
            action=action,
            params=params,
        )

    def install_addon(self, app_id, node_group, addon_settings=None):
        """
        Install an addon.
        """
        # Check if the addon is already installed
        if addon := self.get_installed_addon(app_id=app_id, node_group=node_group):
            logger.info(f"Addon {app_id} is already installed")
            return self.execute_action(
                app_unique_name=addon.get("uniqueName"),
                params=addon_settings,
            )

        logger.info(f"Installing addon {app_id}")
        return self.client.marketplace.App.InstallAddon(
            env_name=self.env_name,
            id=app_id,
            settings=addon_settings,
            node_group=node_group,
        )

    def uninstall_addon(self, app_id, node_group, search=None):
        """
        Uninstall an addon.
        """
        # Check if the addon is already installed
        if addon := self.get_installed_addon(app_id=app_id, node_group=node_group, search=search):
            logger.info(f"Uninstalling addon {app_id}")

            return self.client.marketplace.Installation.Uninstall(
                app_unique_name=addon.get("uniqueName"),
                target_app_id=self.get_env().get("appid"),
                app_template_id=app_id,
            )

        msg = f"Addon {app_id} is not installed on node group {node_group}."
        logger.info(msg)
        return {"result": 0, "message": msg}

    def get_master_node(self, node_group):
        return self.get_nodes(node_group=node_group, is_master=True)

    def get_actions(self):
        actions = self.client.environment.Tracking.GetCurrentActions().get("array", [])
        return [action for action in actions if action.get("text", {}).get("env") == self.env_name]


class PaaSNetBox(PaaS):
    """
    This class is used to manage the PaaS layer of the Jelastic platform for NetBox.
    """

    NETBOX_DIR = '/etc/netbox'

    def get_nb_node_groups(self):
        """
        Get the environment node groups for NetBox.
        """
        results = []

        node_groups = self.get_node_groups()

        # For each node group, get the related nodes
        for node_group in node_groups:
            if "netbox" in node_group.get("node", {}).get("customitem", {}).get("dockerName"):
                results.append(node_group)

        return results

    def load_plugins(self, file_name="plugins.yaml"):
        """
        Loads the plugins from the plugins.yaml file.
        """
        master_node_id = self.get_master_node(NODE_GROUP_CP).get("id")

        try:
            file_path = f"{self.NETBOX_DIR}/config/{file_name}"
            plugins_yaml = self.execute_cmd(master_node_id, f"cat {file_path}")[0].get("out", "")
        except JelasticApiError:
            plugins_yaml = ""

        return yaml.safe_load(plugins_yaml) or {}

    def dump_plugins(self, plugins, file_name="plugins.yaml"):
        """
        Dumps the plugins to the plugins.yaml file.
        """
        master_node_id = self.get_master_node(NODE_GROUP_CP).get("id")
        plugins_yaml = yaml.dump(plugins)

        self.client.environment.File.Write(
            env_name=self.env_name,
            path=f"{self.NETBOX_DIR}/config/{file_name}",
            body=plugins_yaml,
            node_id=master_node_id,
            is_append_mode=False,
        )

    def get_requirements_parser(self):
        master_node_id = self.get_master_node(NODE_GROUP_CP).get('id')
        requirements_str = self.execute_cmd(master_node_id, f'cat {self.NETBOX_DIR}/plugin_requirements.txt')[0].get(
            'out', ''
        )
        return RequirementsParser(requirements_str)

    def write_requirements(self, req: RequirementsParser):
        master_node_id = self.get_master_node(NODE_GROUP_CP).get('id')
        requirements_str = req.dumps()
        return self.client.environment.File.Write(
            env_name=self.env_name,
            node_id=master_node_id,
            body=requirements_str,
            path=f'{self.NETBOX_DIR}/plugin_requirements.txt',
            is_append_mode=False,
        )

    def install_plugin(
        self,
        plugin: dict,
        version,
        plugin_settings=None,
        github_token=None,
        restart: bool = True,
        collectstatic: bool = True,
        install: bool = True,
    ):
        activate_env = "source /opt/netbox/venv/bin/activate"
        # Get all node IDs for the NetBox node groups
        node_ids = [
            node["id"]
            for node_group in self.get_nb_node_groups()
            for node in self.get_nodes(node_group=node_group["name"], is_master=False)
        ]

        # Install the plugin version
        if plugin.get("private"):
            github_url = plugin.get("github_url")
            github_url = github_url.replace("https://github.com", f"git+https://{github_token}@github.com")

            plugin_req = f'{github_url}@{version}'
        else:
            plugin_req = f'{plugin.get("name")}=={version}'

        req = self.get_requirements_parser()
        if req.add(plugin_req):
            self.write_requirements(req)

            if install:
                if plugin.get("private"):
                    # Ensure `git` is installed on every node
                    [self.execute_cmd(node_id, "apt-get install -y git") for node_id in node_ids]

                # Install the plugin
                cmd = f'{activate_env} && pip install -r {self.NETBOX_DIR}/plugin_requirements.txt'
                [self.execute_cmd(node_id, cmd) for node_id in node_ids]

        plugins = self.load_plugins()
        disabled_plugins = self.load_plugins(file_name=DISABLED_PLUGINS_FILE_NAME)

        if plugin.get("app_label") in disabled_plugins:
            disabled_plugins.pop(plugin.get("app_label"))
            self.dump_plugins(disabled_plugins, file_name=DISABLED_PLUGINS_FILE_NAME)

        plugins[plugin.get("app_label")] = plugin_settings or {}
        self.dump_plugins(plugins)

        if collectstatic:
            # Get Node IDs for the CP node group
            cp_node_ids = [node["id"] for node in self.get_nodes(node_group=NODE_GROUP_CP, is_master=False)]
            # Run collectstatic command
            [
                self.execute_cmd(
                    node_id=node_id,
                    command=f"{activate_env} && /opt/netbox/netbox/manage.py collectstatic --no-input --clear 1>/dev/null",
                )
                for node_id in cp_node_ids
            ]

        if restart:
            return self.restart_nodes(
                node_groups=[node_group["name"] for node_group in self.get_nb_node_groups()],
                lazy=True,
            )

        return {"result": 0, "message": "Plugin installed"}

    def uninstall_plugin(self, plugin: dict):
        """
        Uninstall an addon for NetBox.
        """
        # Remove the plugin from the plugins.yaml or disabled_plugins.yaml file
        plugins = self.load_plugins()
        disabled_plugins = self.load_plugins(file_name=DISABLED_PLUGINS_FILE_NAME)

        if plugin.get("app_label") in plugins:
            plugins.pop(plugin.get("app_label"))
            self.dump_plugins(plugins)
        elif plugin.get("app_label") in disabled_plugins:
            disabled_plugins.pop(plugin.get("app_label"))
            self.dump_plugins(disabled_plugins, file_name=DISABLED_PLUGINS_FILE_NAME)

        # Uninstall the plugin from the virtual environment
        master_node_id = self.get_master_node(NODE_GROUP_CP).get("id")
        activate_env = "source /opt/netbox/venv/bin/activate"

        req = self.get_requirements_parser()
        if req.remove(plugin.get("name")):
            self.write_requirements(req)

            # Get all node IDs for the NetBox node groups
            node_ids = [
                node["id"]
                for node_group in self.get_nb_node_groups()
                for node in self.get_nodes(node_group=node_group["name"], is_master=False)
            ]

            # Uninstall the plugin
            cmd = f'{activate_env} && pip uninstall -y {plugin.get("name")}'
            for node_id in node_ids:
                self.execute_cmd(node_id, cmd)

        return self.restart_nodes(
            node_groups=[node_group["name"] for node_group in self.get_nb_node_groups()],
            lazy=True,
        )

    def _enable_disable_plugin(self, plugin: dict, source_file: str, dest_file: str):
        """
        Move a plugin from one file (source) to another (destination).
        """
        # Load source and destination plugin lists
        source_plugins = self.load_plugins(file_name=source_file)
        dest_plugins = self.load_plugins(file_name=dest_file)

        app_label = plugin.get("app_label")
        if app_label in source_plugins:
            # Move the plugin
            dest_plugins[app_label] = source_plugins.pop(app_label)
            # Update the files
            self.dump_plugins(source_plugins, file_name=source_file)
            self.dump_plugins(dest_plugins, file_name=dest_file)

    def disable_plugin(self, plugin: dict):
        """
        Disable a plugin for NetBox.
        """
        self._enable_disable_plugin(plugin, PLUGINS_FILE_NAME, DISABLED_PLUGINS_FILE_NAME)
        return self.restart_nodes(
            node_groups=[node_group["name"] for node_group in self.get_nb_node_groups()],
            lazy=True,
        )

    def enable_plugin(self, plugin: dict):
        """
        Enable a plugin for NetBox.
        """
        self._enable_disable_plugin(plugin, DISABLED_PLUGINS_FILE_NAME, PLUGINS_FILE_NAME)
        return self.restart_nodes(
            node_groups=[node_group["name"] for node_group in self.get_nb_node_groups()],
            lazy=True,
        )

    def get_env_var(self, variable, default=None):
        """
        Get the environment variable for NetBox.
        """
        container_vars = self._get_env_var(NODE_GROUP_CP)
        return getattr(get_config(), variable, container_vars.get(variable, None)) or default

    def _get_docker_tags(self):
        """
        Get the Docker tags for NetBox.
        """
        master_node = self.get_master_node(NODE_GROUP_CP)
        docker = master_node.get("customitem", {})

        if not docker:
            return []

        response = requests.get(f'https://hub.docker.com/v2/repositories/{docker["dockerName"]}/tags?page_size=1000')
        response.raise_for_status()
        response = response.json()

        names = [d["name"] for d in response["results"] if d["name"].startswith("v")]
        tags = []
        for name in names:
            try:
                version = Version.parse(name.lstrip("v"))
                if not version.prerelease:
                    tags.append(version)
            except ValueError:
                pass

        return tags

    def get_upgrades(self, include_patch: bool = False):
        """
        Get the available upgrades for NetBox.
        """
        docker_tags = self._get_docker_tags()
        current_version = Version.parse(settings.VERSION)

        upgrades = []
        for tag in docker_tags:
            if include_patch:
                if tag.major == current_version.major and tag.minor == current_version.minor:
                    upgrades.append(tag)
            if tag > current_version:
                upgrades.append(tag)

        return upgrades

    def is_upgrade_available(self):
        """
        Check if an upgrade is available for NetBox.
        """
        return bool(self.get_upgrades())

    def is_db_backup_running(self, app_unique_name):
        # Get current running actions
        for action in self.get_actions():
            action_parameters = action.get("parameters", {})

            if (
                action_parameters.get("appUniqueName") == app_unique_name
                and action_parameters.get("action") == "backup"
            ):
                return True

        return False

    def db_backup(self, app_unique_name):
        """
        Backup NetBox database.
        """
        while self.is_db_backup_running(app_unique_name=app_unique_name):
            logger.debug("Waiting for database backup to finish...")
            time.sleep(30)

        return self.execute_action(app_unique_name=app_unique_name, action="backup")

    def upgrade_checks(self, version) -> Tuple[bool, str]:
        """
        Run upgrade checks for NetBox.
        """
        # Fetch the plugins from the store
        plugins = utils.get_plugins_list()
        for plugin_name, plugin in plugins.items():
            if plugin.get("app_label") in settings.PLUGINS:
                if not utils.filter_releases(plugin, version):
                    return (
                        False,
                        f"Plugin {plugin_name} does not have a release for version {version}",
                    )

        return True, ""

    def upgrade(self, version, lic=None):
        """
        Upgrade NetBox.
        """
        activate_env = "source /opt/netbox/venv/bin/activate"
        master_node_id = self.get_master_node(NODE_GROUP_CP).get("id")
        env_app_id = self.get_env().get("appid")

        if addon := self.get_installed_addon(app_id="db-backup", node_group=NODE_GROUP_SQLDB):
            self.db_backup(app_unique_name=addon.get("uniqueName"))

        # Upgrade all plugins to the latest compatible version first without restarting the nodes
        for plugin_name, plugin in utils.get_plugins_list().items():
            if plugin.get('app_label') in settings.PLUGINS:
                plugin_version = utils.filter_releases(plugin, version)[0]
                self.install_plugin(
                    plugin=plugin,
                    version=plugin_version,
                    plugin_settings=settings.PLUGINS_CONFIG.get(plugin.get('app_label'), {}),
                    github_token=lic,
                    restart=False,
                    collectstatic=False,  # Do not run during upgrade as it may crash due to version incompatibility
                    install=False,  # Do not install the plugin as it will be installed during the upgrade
                )

        logger.info(f"Upgrading NetBox {NODE_GROUP_CP} nodes to version v{version}")
        self.client.environment.Control.RedeployContainersByGroup(
            env_name=self.env_name,
            node_group=NODE_GROUP_CP,
            tag=f"v{version}",
            use_existing_volumes=True,
            manage_dns_state=True,
            is_sequential=True,
            delay=60000,
        )

        # Run collectstatic command
        self.execute_cmd(
            node_id=master_node_id,
            command=f"{activate_env} && /opt/netbox/netbox/manage.py collectstatic --no-input --clear 1>/dev/null",
        )

        # Get all node groups except the CP node group
        node_groups = [
            node_group["name"] for node_group in self.get_nb_node_groups() if node_group["name"] != NODE_GROUP_CP
        ]
        redeploy_script = """
        import com.hivext.api.Response;

        function checkEnvStatus() {
            let retries = 0, waitInterval = (interval * 1000), resp, c = jelastic.environment.control, e = envName, s = session, r
            do {
                java.lang.Thread.sleep(retries ? waitInterval : 0)
                resp = api.env.control.GetEnvInfo({ envName: envName, lazy: true })
                if (resp.result != 0 && resp.result != Response.APPLICATION_NOT_EXIST) return resp;
                api.marketplace.console.WriteLog(logPrefix + " [" + (retries + 1) + "/"+ maxRetries + "] checking " + envName + " status result: " + (resp.env ? resp.env.status : resp ));
            } while (resp.result != Response.APPLICATION_NOT_EXIST && resp.env.status != 1 && (++retries < maxRetries))
        }

        function checkCurrentActions() {
            let retries = 0, waitInterval = (interval * 1000), resp, c = jelastic.environment.tracking, e = envName, s = session, r
            do {
                java.lang.Thread.sleep(retries ? waitInterval: 0)
                resp = api.env.tracking.GetCurrentActions()
                if (resp.result != 0) return resp;
                var currentActions = resp.array.filter(function(action) {
                    return action.text && action.text.env == envName
                })
                if (!currentActions.length) break;
                api.marketplace.console.WriteLog(logPrefix + " [" + (retries + 1) + "/"+ maxRetries + "] waiting for " + envName + " actions to finish");
            } while (++retries < maxRetries)
        }

        checkEnvStatus();
        checkCurrentActions();

        api.marketplace.console.WriteLog("RedeployContainers: " + nodeGroup);
        return api.env.control.RedeployContainersByGroup({ envName: envName, session: session, nodeGroup: nodeGroup, tag: tag, useExistingVolumes: true, manageDnsState: true, isSequential: true });
        """

        batch_request = {
            "alias": {"m": "Dev.Scripting.EvalCode"},
            "global": {"appid": env_app_id, "session": self.client._token, "type": "js", "code": redeploy_script},
            "methods": [],
        }

        for node_group in node_groups:
            batch_request["methods"].append(
                {
                    "m": {
                        "params": {
                            "envName": self.env_name,
                            "nodeGroup": node_group,
                            "tag": f"v{version}",
                            "maxRetries": 100,
                            "interval": 60,
                            "logPrefix": node_group,
                        }
                    }
                }
            )

        batch_code = f"""
        let resp = api.utils.batch.Call(appid, toJSON({batch_request}), true);
        if (resp.result != 0) return resp;
        """

        self.run_script(
            name="ncp-upgrade-netbox",
            description=f"Upgrading NetBox nodes to version v{version}",
            code=batch_code,
            delay=10,
        )



from django import forms
from django.contrib.postgres.forms.array import SimpleArrayField

from extras.constants import DEFAULT_DASHBOARD
from utilities.forms import BOOLEAN_WITH_BLANK_CHOICES
from .nb_settings import *
from .validators import validate_comma_separated

JELASTIC_API = "https://app.xapp.cloudmydc.com"
NETBOX_JPS_REPO = "https://raw.githubusercontent.com/Onemind-Services-LLC/netbox-jps/master"
NODE_GROUP_CP = "cp"
NODE_GROUP_SQLDB = "sqldb"

PLUGINS_FILE_NAME = "plugins.yaml"
DISABLED_PLUGINS_FILE_NAME = "disabled_plugins.yaml"

SECTION_REQUIRED = None
SECTION_SYSTEM = "System"
SECTION_SECURITY = "Security"
SECTION_DATA = "Data"
SECTION_DEFAULT = "Default"
SECTION_DATETIME = "Date/Time"
SECTION_MISC = "Miscellaneous"
SECTION_DEVELOPMENT = "Development"
SECTION_CONTAINER = "Container"

NETBOX_SETTINGS = NbSettings(
    [
        NbSection(
            name=SECTION_REQUIRED,
            params=[
                Param(
                    key="ALLOWED_HOSTS",
                    label="Allowed Hosts",
                    help_text="Comma separated list of allowed hosts (FQDN, IP address, or pattern), or '*' for all",
                    placeholder="netbox.example.com, localhost, *",
                    field=forms.CharField,
                    field_kwargs={"validators": [validate_comma_separated]},
                ),
                Param(
                    key="DB_CONN_MAX_AGE",
                    label="DB Connection Max Age",
                    help_text="Maximum age of database connections, in seconds",
                    placeholder="300",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
                Param(
                    key="DB_DISABLE_SERVER_SIDE_CURSORS",
                    label="DB Disable Server Side Cursors",
                    help_text="Disable server-side cursors",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="DB_HOST",
                    label="DB Host",
                    help_text="Database server hostname or IP address",
                    placeholder="localhost",
                ),
                Param(
                    key="DB_NAME",
                    label="DB Name",
                    help_text="Database name",
                    placeholder="netbox",
                ),
                Param(
                    key="DB_PASSWORD",
                    label="DB Password",
                    help_text="Database user password",
                    placeholder="password",
                ),
                Param(
                    key="DB_PORT",
                    label="DB Port",
                    help_text="Database server port",
                    placeholder="5432",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                    initial=5432,
                ),
                Param(
                    key="DB_SSLMODE",
                    label="DB SSL Mode",
                    help_text="Database SSL mode",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": [
                            ("disable", "Disable"),
                            ("allow", "Allow"),
                            ("prefer", "Prefer"),
                            ("require", "Require"),
                            ("verify-ca", "Verify CA"),
                            ("verify-full", "Verify Full"),
                        ],
                    },
                    initial="prefer",
                ),
                Param(
                    key="DB_USER",
                    label="DB User",
                    help_text="Database username",
                    placeholder="netbox",
                ),
                Param(
                    key="DB_WAIT_TIMEOUT",
                    label="DB Wait Timeout",
                    help_text="Database connection wait timeout, in seconds",
                    placeholder="300",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                    initial=300,
                ),
                Param(
                    key="REDIS_HOST",
                    label="Redis Host",
                    help_text="Redis server hostname or IP address",
                    placeholder="redis",
                ),
                Param(
                    key="REDIS_INSECURE_SKIP_TLS_VERIFY",
                    label="Redis Insecure Skip TLS Verify",
                    help_text="Skip TLS verification when connecting to Redis",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="REDIS_PASSWORD",
                    label="Redis Password",
                    help_text="Redis server password",
                    placeholder="password",
                    required=False,
                ),
                Param(
                    key="REDIS_PORT",
                    label="Redis Port",
                    help_text="Redis server port",
                    placeholder="6379",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                    initial=6379,
                ),
                Param(
                    key="REDIS_SSL",
                    label="Redis SSL",
                    help_text="Use SSL when connecting to Redis",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="REDIS_USERNAME",
                    label="Redis Username",
                    help_text="Redis server username",
                    placeholder="redis",
                    required=False,
                ),
            ],
        ),
        NbSection(
            name=SECTION_SYSTEM,
            params=[
                Param(
                    key="BASE_PATH",
                    label="Base Path",
                    help_text="Base path for URL patterns",
                    placeholder="/",
                    required=False,
                    initial=None,
                ),
                Param(
                    key="EMAIL_FROM",
                    label="Email From",
                    help_text="Default email sender address",
                    placeholder="NetBox <netbox@example.com>",
                    required=False,
                ),
                Param(
                    key="EMAIL_PASSWORD",
                    label="Email Password",
                    help_text="SMTP server password",
                    placeholder="password",
                    required=False,
                ),
                Param(
                    key="EMAIL_PORT",
                    label="Email Port",
                    help_text="SMTP server port",
                    placeholder="25",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
                Param(
                    key="EMAIL_SERVER",
                    label="Email Server",
                    help_text="SMTP server hostname or IP address",
                    placeholder="localhost",
                    required=False,
                ),
                Param(
                    key="EMAIL_SSL_CERTFILE",
                    label="Email SSL Certfile",
                    help_text="SMTP SSL certificate file",
                    placeholder="/path/to/certfile",
                    required=False,
                ),
                Param(
                    key="EMAIL_SSL_KEYFILE",
                    label="Email SSL Keyfile",
                    help_text="SMTP SSL key file",
                    placeholder="/path/to/keyfile",
                    required=False,
                ),
                Param(
                    key="EMAIL_TIMEOUT",
                    label="Email Timeout",
                    help_text="SMTP server timeout, in seconds",
                    placeholder="10",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
                Param(
                    key="EMAIL_USERNAME",
                    label="Email Username",
                    help_text="SMTP server username",
                    placeholder="admin@example.com",
                    required=False,
                    field_kwargs={"widget": forms.EmailInput()},
                ),
                Param(
                    key="EMAIL_USE_SSL",
                    label="Email Use SSL",
                    help_text="Use SSL when connecting to SMTP server",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="EMAIL_USE_TLS",
                    label="Email Use TLS",
                    help_text="Use TLS when connecting to SMTP server",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="ENABLE_LOCALIZATION",
                    label="Enable Localization",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="HTTP_PROXIES",
                    label="HTTP Proxies",
                    help_text="Dictionary of HTTP proxies",
                    placeholder="{'http': 'http://proxy.example.com:3128'}",
                    required=False,
                ),
                Param(
                    key="INTERNAL_IPS",
                    label="Internal IPs",
                    help_text="List of internal IP addresses",
                    placeholder="127.0.0.1, ::1",
                    required=False,
                ),
                Param(
                    key="JINJA2_FILTERS",
                    label="Jinja2 Filters",
                    help_text="Dictionary of Jinja2 filters",
                    placeholder="{'filter_name': 'path.to.filter'}",
                    required=False,
                    field=forms.JSONField,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                    initial={},
                ),
                Param(
                    key="MEDIA_ROOT",
                    label="Media Root",
                    help_text="Absolute filesystem path to media files",
                    placeholder="/opt/netbox/netbox/media",
                    required=False,
                ),
                Param(
                    key="REPORTS_ROOT",
                    label="Reports Root",
                    help_text="Absolute filesystem path to reports",
                    placeholder="/opt/netbox/netbox/reports",
                    required=False,
                ),
                Param(
                    key="SCRIPTS_ROOT",
                    label="Scripts Root",
                    help_text="Absolute filesystem path to custom scripts",
                    placeholder="/opt/netbox/netbox/scripts",
                    required=False,
                ),
                Param(
                    key="STORAGE_BACKEND",
                    label="Storage Backend",
                    help_text="Storage backend for file attachments",
                    placeholder="django.core.files.storage.FileSystemStorage",
                    required=False,
                ),
                Param(
                    key="STORAGE_CONFIG",
                    label="Storage Config",
                    help_text="Configuration parameters for the storage backend",
                    placeholder="{'location': '/var/netbox/media'}",
                    required=False,
                    field=forms.JSONField,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                    initial={},
                ),
            ],
        ),
        NbSection(
            name=SECTION_SECURITY,
            params=[
                Param(
                    key="ALLOW_TOKEN_RETRIEVAL",
                    label="Allow Token Retrieval",
                    help_text="Allow retrieval of user tokens in API or UI",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                    initial=True,
                ),
                Param(
                    key="ALLOWED_URL_SCHEMES",
                    label="Allowed URL Schemes",
                    help_text="List of allowed URL schemes",
                    placeholder="http, https",
                    required=False,
                    field=SimpleArrayField,
                    field_kwargs={'base_field': forms.CharField()},
                ),
                Param(
                    key="AUTH_PASSWORD_VALIDATORS",
                    label="Auth Password Validators",
                    help_text="List of password validators",
                    placeholder="[]",
                    required=False,
                    field=forms.JSONField,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                    initial=[
                        {
                            "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
                        },
                        {
                            "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
                            "OPTIONS": {
                                "min_length": 10,
                            },
                        },
                        {
                            "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
                        },
                        {
                            "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
                        },
                    ],
                ),
                Param(
                    key="CORS_ORIGIN_ALLOW_ALL",
                    label="CORS Origin Allow All",
                    help_text="Allow all CORS origins",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="CORS_ORIGIN_REGEX_WHITELIST",
                    label="CORS Origin Regex Whitelist",
                    help_text="List of CORS origin regex patterns",
                    placeholder="https://*.example.com, https://hostname.example.com",
                    required=False,
                    field=SimpleArrayField,
                    field_kwargs={'base_field': forms.CharField()},
                ),
                Param(
                    key="CORS_ORIGIN_WHITELIST",
                    label="CORS Origin Whitelist",
                    help_text="List of CORS origins",
                    placeholder="https://*.example.com, https://hostname.example.com",
                    required=False,
                    field=SimpleArrayField,
                    field_kwargs={'base_field': forms.CharField()},
                ),
                Param(
                    key="CSRF_COOKIE_NAME",
                    label="CSRF Cookie Name",
                    help_text="Name of the CSRF cookie",
                    placeholder="csrftoken",
                    required=False,
                    initial="csrftoken",
                ),
                Param(
                    key="CSRF_COOKIE_SECURE",
                    label="CSRF Cookie Secure",
                    help_text="Require HTTPS for CSRF cookie",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="CSRF_TRUSTED_ORIGINS",
                    label="CSRF Trusted Origins",
                    help_text="List of trusted origins for CSRF",
                    placeholder="https://*.example.com, https://hostname.example.com",
                    required=False,
                    field=SimpleArrayField,
                    field_kwargs={'base_field': forms.CharField()},
                ),
                Param(
                    key="LOGIN_PERSISTENCE",
                    label="Login Persistence",
                    help_text="Enable session persistence after browser restart",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                    initial=True,
                ),
                Param(
                    key="LOGIN_REQUIRED",
                    label="Login Required",
                    help_text="Require authentication to access any data",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="LOGIN_TIMEOUT",
                    label="Login Timeout",
                    help_text="Session timeout, in seconds",
                    placeholder="1209600",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                    initial=1209600,
                ),
                Param(
                    key="SECURE_SSL_REDIRECT",
                    label="Secure SSL Redirect",
                    help_text="Redirect all non-HTTPS requests to HTTPS",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
            ],
        ),
        NbSection(
            name=SECTION_DATA,
            params=[
                Param(
                    key="CUSTOM_VALIDATORS",
                    label="Custom Validators",
                    help_text="List of custom data validators",
                    placeholder="[]",
                    required=False,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                ),
                Param(
                    key="FIELD_CHOICES",
                    label="Field Choices",
                    help_text="List of custom field choices",
                    placeholder="[]",
                    required=False,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                ),
            ],
        ),
        NbSection(
            name=SECTION_DEFAULT,
            params=[
                Param(
                    key="DEFAULT_DASHBOARD",
                    label="Default Dashboard",
                    help_text="Default dashboard for users",
                    required=False,
                    field=forms.JSONField,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                    initial=DEFAULT_DASHBOARD,
                ),
                Param(
                    key="DEFAULT_USER_PREFERENCES",
                    label="Default User Preferences",
                    help_text="Default user preferences",
                    required=False,
                    field=forms.JSONField,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                    initial={},
                ),
                Param(
                    key="PAGINATE_COUNT",
                    label="Paginate Count",
                    help_text="Default number of objects to display per page",
                    placeholder="50",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 1})},
                ),
                Param(
                    key="POWERFEED_DEFAULT_AMPERAGE",
                    label="Powerfeed Default Amperage",
                    help_text="Default power feed amperage",
                    placeholder="20",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
                Param(
                    key="POWERFEED_DEFAULT_MAX_UTILIZATION",
                    label="Powerfeed Default Max Utilization",
                    help_text="Default power feed maximum utilization",
                    placeholder="80",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
                Param(
                    key="POWERFEED_DEFAULT_VOLTAGE",
                    label="Powerfeed Default Voltage",
                    help_text="Default power feed voltage",
                    placeholder="120",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
                Param(
                    key="RACK_ELEVATION_DEFAULT_UNIT_HEIGHT",
                    label="Rack Elevation Default Unit Height",
                    help_text="Default rack elevation unit height",
                    placeholder="44",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
                Param(
                    key="RACK_ELEVATION_DEFAULT_UNIT_WIDTH",
                    label="Rack Elevation Default Unit Width",
                    help_text="Default rack elevation unit width",
                    placeholder="50",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
            ],
        ),
        NbSection(
            name=SECTION_DATETIME,
            params=[
                Param(
                    key="DATETIME_FORMAT",
                    label="Datetime Format",
                    help_text="Default datetime format",
                    placeholder="N j, Y, H:i:s T",
                    required=False,
                ),
                Param(
                    key="DATE_FORMAT",
                    label="Date Format",
                    help_text="Default date format",
                    placeholder="N j, Y",
                    required=False,
                ),
                Param(
                    key="SHORT_DATETIME_FORMAT",
                    label="Short Datetime Format",
                    help_text="Default short datetime format",
                    placeholder="m/d/Y P",
                    required=False,
                ),
                Param(
                    key="SHORT_DATE_FORMAT",
                    label="Short Date Format",
                    help_text="Default short date format",
                    placeholder="m/d/Y",
                    required=False,
                ),
                Param(
                    key="SHORT_TIME_FORMAT",
                    label="Short Time Format",
                    help_text="Default short time format",
                    placeholder="H:i:s",
                    required=False,
                ),
                Param(
                    key="TIME_FORMAT",
                    label="Time Format",
                    help_text="Default time format",
                    placeholder="H:i:s",
                    required=False,
                ),
                Param(
                    key="TIME_ZONE",
                    label="Time Zone",
                    help_text="Default time zone",
                    placeholder="UTC",
                    required=False,
                ),
            ],
        ),
        NbSection(
            name=SECTION_MISC,
            params=[
                Param(
                    key="ADMINS",
                    label="Admins",
                    help_text="List of NetBox administrators",
                    placeholder="[('NetBox Admin', 'admin@example.com')]",
                    required=False,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                ),
                Param(
                    key="BANNER_BOTTOM",
                    label="Banner Bottom",
                    help_text="Bottom banner text",
                    placeholder="Banner text",
                    required=False,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                ),
                Param(
                    key="BANNER_LOGIN",
                    label="Banner Login",
                    help_text="Login banner text",
                    placeholder="Banner text",
                    required=False,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                ),
                Param(
                    key="BANNER_TOP",
                    label="Banner Top",
                    help_text="Top banner text",
                    placeholder="Banner text",
                    required=False,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                ),
                Param(
                    key="BANNER_MAINTENANCE",
                    label="Banner Maintenance",
                    help_text="Maintenance banner text",
                    placeholder="Banner text",
                    required=False,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                ),
                Param(
                    key="CENSUS_REPORTING_ENABLED",
                    label="Census Reporting Enabled",
                    help_text="Enable census reporting",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="CHANGELOG_RETENTION",
                    label="Changelog Retention",
                    help_text="Change log retention policy, in days",
                    placeholder="90",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
                Param(
                    key="DATA_UPLOAD_MAX_MEMORY_SIZE",
                    label="Data Upload Max Memory Size",
                    help_text="Maximum size of file uploads, in bytes",
                    placeholder="2621440",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
                Param(
                    key="ENFORCE_GLOBAL_UNIQUE",
                    label="Enforce Global Unique",
                    help_text="Enforce global uniqueness of IP addresses",
                    placeholder="False",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="FILE_UPLOAD_MAX_MEMORY_SIZE",
                    label="File Upload Max Memory Size",
                    help_text="Maximum size of file uploads, in bytes",
                    placeholder="2621440",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
                Param(
                    key="GRAPHQL_ENABLED",
                    label="GraphQL Enabled",
                    help_text="Enable GraphQL API",
                    placeholder="False",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="JOB_RETENTION",
                    label="Job Retention",
                    help_text="Job retention policy, in days",
                    placeholder="365",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
                Param(
                    key="MAINTENANCE_MODE",
                    label="Maintenance Mode",
                    help_text="Enable maintenance mode",
                    placeholder="False",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="MAPS_URL",
                    label="Maps URL",
                    help_text="URL to link to external maps",
                    placeholder="https://maps.google.com/?q=",
                    required=False,
                    field=forms.URLField,
                ),
                Param(
                    key="MAX_PAGE_SIZE",
                    label="Max Page Size",
                    help_text="Maximum page size for paginated results",
                    placeholder="1000",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 1})},
                ),
                Param(
                    key="METRICS_ENABLED",
                    label="Metrics Enabled",
                    help_text="Enable metrics reporting",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="PREFER_IPV4",
                    label="Prefer IPv4",
                    help_text="Prefer IPv4 addresses over IPv6",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="QUEUE_MAPPINGS",
                    label="Queue Mappings",
                    help_text="Dictionary of queue mappings",
                    placeholder="{'queue_name': 'queue_name'}",
                    required=False,
                    field=forms.JSONField,
                    field_kwargs={"widget": forms.Textarea(attrs={"class": "vLargeTextField"})},
                ),
                Param(
                    key="RELEASE_CHECK_URL",
                    label="Release Check URL",
                    help_text="URL for release check",
                    placeholder="https://api.github.com/repos/netbox-community/netbox/releases",
                    required=False,
                    field=forms.URLField,
                ),
                Param(
                    key="RQ_DEFAULT_TIMEOUT",
                    label="RQ Default Timeout",
                    help_text="Default timeout for RQ jobs, in seconds",
                    placeholder="300",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 1})},
                ),
                Param(
                    key="RQ_RETRY_INTERVAL",
                    label="RQ Retry Interval",
                    help_text="Interval between RQ job retries, in seconds",
                    placeholder="60",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 1})},
                ),
                Param(
                    key="RQ_RETRY_MAX",
                    label="RQ Retry Max",
                    help_text="Maximum number of RQ job retries",
                    placeholder="3",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 0})},
                ),
            ],
        ),
        NbSection(
            name=SECTION_DEVELOPMENT,
            params=[
                Param(
                    key="DEBUG",
                    label="Debug",
                    help_text="Enable debugging mode",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
                Param(
                    key="DEVELOPER",
                    label="Developer",
                    help_text="Enable developer mode, use with caution",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": BOOLEAN_WITH_BLANK_CHOICES,
                    },
                ),
            ],
        ),
        NbSection(
            name=SECTION_CONTAINER,
            params=[
                Param(
                    key="HOUSEKEEPING_INTERVAL",
                    label="Housekeeping Interval",
                    help_text="Interval between housekeeping tasks, in seconds",
                    placeholder="3600",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 10})},
                ),
                Param(
                    key="LOGLEVEL",
                    label="Loglevel",
                    help_text="Log level",
                    placeholder="INFO",
                    required=False,
                    field=forms.ChoiceField,
                    field_kwargs={
                        "widget": forms.Select(),
                        "choices": [
                            ("DEBUG", "DEBUG"),
                            ("INFO", "INFO"),
                            ("WARNING", "WARNING"),
                            ("ERROR", "ERROR"),
                            ("CRITICAL", "CRITICAL"),
                        ],
                    },
                ),
                Param(
                    key="MAX_DB_WAIT_TIME",
                    label="Max DB Wait Time",
                    help_text="Maximum time to wait for database, in seconds",
                    placeholder="300",
                    required=False,
                    field_kwargs={"widget": forms.NumberInput(attrs={"min": 1})},
                ),
            ],
        ),
    ]
)

NETBOX_SUPERUSER_SETTINGS = ["SUPERUSER_NAME", "SUPERUSER_EMAIL", "SUPERUSER_PASSWORD"]



import os

from setuptools import find_packages, setup

readme = os.path.join(os.path.dirname(__file__), 'README.md')

with open(readme) as fh:
    long_description = fh.read()

setup(
    name="netbox-paas",
    version="0.3.6",
    description="Enhances NetBox on CloudMyDC's PaaS with advanced management and control features.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Onemind-Services-LLC/netbox-paas/",
    download_url="https://github.com/Onemind-Services-LLC/netbox-paas/",
    author="Abhimanyu Saharan",
    author_email="asaharan@onemindservices.com",
    maintainer="Abhimanyu Saharan",
    maintainer_email="asaharan@onemindservices.com",
    install_requires=["py-jelastic", "croniter<3.0.0", "semver<4.0.0", "requirements-parser", "elastic-apm"],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    platform=[],
    keywords=["netbox", "netbox-plugin"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Django",
        "Programming Language :: Python :: 3",
    ],
)



import django_tables2 as tables
from netbox.tables import columns


__all__ = ("NetBoxBackupsTable",)

RESTORE_BUTTON = """
<form method="post" action="{% url 'plugins:netbox_paas:netboxdbbackup_restore' pk=object.pk %}">
  {% csrf_token %}
  <input type="hidden" name="name" value="{{ record.name }}">
  <button type="submit" class="btn btn-sm btn-warning">
    <i class="mdi mdi-restore"></i> Restore
  </button>
</form>
"""


class NetBoxBackupsTable(tables.Table):
    name = tables.Column(
        orderable=False,
    )
    datetime = columns.DateTimeColumn(
        orderable=False,
        verbose_name="Date/Time",
    )
    timezone = tables.Column(
        orderable=False,
    )

    backup_type = columns.TemplateColumn(
        template_code="""{{ record.backup_type|title }}""",
        orderable=False,
        verbose_name="Backup Type",
    )

    db_version = tables.Column(
        orderable=False,
        verbose_name="Database Version",
    )

    actions = columns.TemplateColumn(
        template_code=RESTORE_BUTTON,
        orderable=False,
        verbose_name="Actions",
    )

    class Meta:
        fields = ("name", "datetime", "timezone", "backup_type", "db_version", "actions")
        attrs = {
            "class": "table table-hover object-list",
        }



import requests
from django import forms
from django.conf import settings
from django.forms import ValidationError
from django.utils.translation import gettext as _

from netbox.forms import NetBoxModelForm
from utilities.forms import BootstrapMixin, ConfirmationForm as _ConfirmationForm
from utilities.forms.fields import CommentField
from utilities.rqworker import get_workers_for_queue
from .constants import NETBOX_SETTINGS, NODE_GROUP_SQLDB, DISABLED_PLUGINS_FILE_NAME
from .models import *
from .utils import *

__all__ = (
    "NetBoxConfigurationForm",
    "NetBoxSettingsForm",
    "NetBoxBackupStorageForm",
    "NetBoxDBBackupForm",
    "NetBoxPluginInstallForm",
)


def create_fieldset():
    """
    Create a fieldset from the NETBOX_SETTINGS.
    """
    fieldset = ()

    # Iterate through each section in NETBOX_SETTINGS
    for section in NETBOX_SETTINGS.sections:
        fields = []

        # Iterate through each setting in the section
        for param in section.params:
            # Append the lowercase name of the setting to the fields list
            fields.append(param.key.lower())

        # Append the section name and its fields to the fieldset
        fieldset += ((section.name, fields),)

    return fieldset


class NetBoxConfigurationForm(NetBoxModelForm):
    key = forms.CharField(
        help_text="Jelastic API token where the NetBox instance is running.",
    )

    env_name_storage = forms.ChoiceField(
        label="Environment Name",
        required=False,
        help_text="Jelastic environment name where the backup storage is running.",
    )

    license = forms.CharField(
        required=False,
        help_text="NetBox Enterprise license key.",
    )

    comments = CommentField()

    fieldsets = (
        (None, ("key", "description")),
        (_("Backup"), ("env_name_storage",)),
        (_("Enterprise"), ("license",)),
    )

    class Meta:
        model = NetBoxConfiguration
        fields = (
            "key",
            "description",
            "env_name_storage",
            "license",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Disable env_name_storage if key is empty
        if not self.instance.key:
            self.fields["env_name_storage"].disabled = True
        else:
            self.fields["env_name_storage"].initial = self.instance.env_name_storage

            env_infos = (
                self.instance.paas(self.instance.env_name, auto_init=False)
                .client.environment.Control.GetEnvs()
                .get("infos", [])
            )

            # Fetch the environment list and build the choices
            self.fields["env_name_storage"].choices = [
                (
                    env_info["env"]["envName"],
                    f"{env_info['env']['displayName']} ({env_info['env']['envName']})",
                )
                for env_info in env_infos
                if env_info.get("env", {}).get("properties", {}).get("projectScope", "") == "backup"
            ]


class NetBoxSettingsForm(BootstrapMixin, forms.Form):
    fieldsets = create_fieldset()

    class Meta:
        fields = [param.key.lower() for section in NETBOX_SETTINGS.sections for param in section.params]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Dynamically create fields for each setting
        for _, params in kwargs.get("initial", {}).items():
            for param in params:
                self.fields[param.key.lower()] = param.field(
                    label=param.label,
                    required=param.required,
                    help_text=param.help_text,
                    initial=param.initial,
                    **param.field_kwargs,
                )

    def clean(self):
        cleaned_data = super().clean()

        if not get_workers_for_queue('default'):
            raise ValidationError(
                "No RQ workers operating on the 'default' queue are currently active in the environment."
            )

        instance = NetBoxConfiguration.objects.first()
        env = instance.get_env()

        # Ensure no actions are currently running on the environment
        if env.get_actions():
            raise ValidationError("There are currently actions running on the environment.")

        data = {}

        # Iterate through each section in NETBOX_SETTINGS
        for section in NETBOX_SETTINGS.sections:
            data[section.name] = {}

            # Iterate through each setting in the section
            for param in section.params:
                # Get the value of the setting
                value = cleaned_data.get(param.key.lower())

                # Set the value of the setting
                data[section.name][param.key.upper()] = value

        return data


class NetBoxBackupStorageForm(BootstrapMixin, forms.Form):
    deployment = forms.ChoiceField(
        choices=(
            ("standalone", "Standalone"),
            ("cluster", "Cluster"),
        )
    )

    node_count = forms.ChoiceField(
        choices=((1, 1), (3, 3), (5, 5), (7, 7)),
        initial=1,
        help_text="Number of nodes in the cluster.",
        required=False,
    )

    storage_size = forms.IntegerField(
        label="Storage Size",
        required=False,
        help_text="Size of the storage in GB.",
        initial=10,
        max_value=200,
    )

    region = forms.ChoiceField(
        help_text="Region where the storage will be deployed, this can be different from the NetBox instance region."
    )

    display_name = forms.CharField(
        label="Display Name",
        help_text="Display name for the storage.",
        max_length=50,
        initial="Backup Storage",
    )

    fieldsets = (
        (None, ("display_name", "storage_size", "region")),
        ("Deployment", ("deployment", "node_count")),
    )

    class Meta:
        fields = [
            "deployment",
            "node_count",
            "storage_size",
            "region",
            "display_name",
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Fetch the region list and build the choices
        nc = NetBoxConfiguration.objects.first()
        regions = nc.paas(nc.env_name, auto_init=False).client.environment.Control.GetRegions().get("array", [])
        self.fields["region"].choices = [
            (hard_node_group["uniqueName"], region["displayName"])
            for region in regions
            for hard_node_group in region["hardNodeGroups"]
        ]

    def clean(self):
        cleaned_data = super().clean()

        if not get_workers_for_queue('default'):
            raise ValidationError(
                "No RQ workers operating on the 'default' queue are currently active in the environment."
            )

        if cleaned_data["deployment"] == "cluster":
            if cleaned_data["node_count"] == "1":
                raise ValidationError({"node_count": "Node count must be greater than 1 for a cluster deployment."})

        if cleaned_data["deployment"] == "standalone":
            if cleaned_data["node_count"] != "1":
                raise ValidationError({"node_count": "Node count must be 1 for a standalone deployment."})

        return cleaned_data


class NetBoxDBBackupForm(NetBoxModelForm):
    db_password = forms.CharField(
        label="Database Password",
        help_text="Password for the <strong>webadmin</strong> user. You will find this in your email.",
        required=False,
    )

    fieldsets = (
        (None, ("netbox_env", "db_password")),
        ("Backup", ("crontab", "keep_backups")),
    )

    class Meta:
        model = NetBoxDBBackup
        fields = ["netbox_env", "crontab", "keep_backups", "db_password"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.instance.pk:
            # Fetch the database password from the addon settings
            app = self.instance.netbox_env.get_env().get_installed_addon("db-backup", node_group=NODE_GROUP_SQLDB)
            data = app.get("settings", {}).get("main", {}).get("data", {})
            self.fields["db_password"].initial = data.get("dbpass")

    def clean(self):
        super().clean()

        if not get_workers_for_queue('default'):
            raise ValidationError(
                "No RQ workers operating on the 'default' queue are currently active in the environment."
            )

        if db_password := self.cleaned_data.get("db_password"):
            self.instance._db_password = db_password

        if not self.instance.pk and not db_password:
            raise ValidationError({"db_password": "This field is required when adding a new backup."})


class NetBoxPluginInstallForm(BootstrapMixin, forms.Form):
    name = forms.CharField(
        label="Plugin Name",
        help_text="Name of the plugin to install.",
        max_length=255,
        disabled=True,
    )

    version = forms.ChoiceField(
        label="Plugin Version",
        help_text="Version of the plugin to install.",
    )

    configuration = forms.JSONField(
        label="Configuration",
        help_text="Configuration for the plugin.",
        required=False,
        initial={},
    )

    fieldsets = (
        (None, ("name", "version")),
        (_("Configuration"), ("configuration",)),
    )

    class Meta:
        fields = ["name", "version", "configuration"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        initial = kwargs.get("initial", {})

        # Get the plugins.yaml
        plugins = get_plugins_list()
        if plugin := plugins.get(initial.get("name")):
            self.fields["version"].choices = [(release, release) for release in filter_releases(plugin)]

        if initial.get("type") == "update":
            from importlib.metadata import metadata

            plugin_name = plugin.get("app_label")
            self.fields["version"].initial = metadata(plugin_name).get("Version")

            if settings.PLUGINS_CONFIG.get(plugin_name):
                self.fields["configuration"].initial = settings.PLUGINS_CONFIG[plugin_name]
            elif (
                disabled_plugins := NetBoxConfiguration.objects.first()
                .get_env()
                .load_plugins(DISABLED_PLUGINS_FILE_NAME)
            ):
                self.fields["configuration"].initial = disabled_plugins.get(plugin_name, {})

    def clean(self):
        if not get_workers_for_queue('default'):
            raise ValidationError(
                "No RQ workers operating on the 'default' queue are currently active in the environment."
            )

        instance = NetBoxConfiguration.objects.first()
        env = instance.get_env()

        # Ensure no actions are currently running on the environment
        if env.get_actions():
            raise ValidationError("There are currently actions running on the environment.")

        plugins = get_plugins_list()
        plugin = plugins.get(self.cleaned_data.get("name"))
        selected_version = self.cleaned_data.get("version")

        # If the plugin is private, ensure that license is provided
        nc = NetBoxConfiguration.objects.first()
        if plugin.get("private"):
            if not nc.license:
                raise ValidationError({"name": "This plugin requires a NetBox Enterprise license."})

            # Check if the plugin is accessible using the license
            response = requests.get(
                plugin.get("github_api_url"),
                headers={"Authorization": f"Bearer {nc.license}"},
            )
            if not response.ok:
                raise ValidationError({"name": "This plugin is not accessible using the provided license."})

        # Get the required_settings from the plugin
        required_settings = next(
            (
                release.get("netbox", {}).get("required_settings", [])
                for release in plugin.get("releases", [])
                if release.get("tag") == selected_version
            ),
            [],
        )
        configuration = self.cleaned_data.get("configuration")

        # Check if the required_settings are in the configuration
        if not all(key in configuration.keys() for key in required_settings):
            raise ValidationError({"configuration": f"Missing required settings: {', '.join(required_settings)}"})


class NetBoxUpgradeForm(BootstrapMixin, forms.Form):
    version = forms.ChoiceField(
        label="Version",
        help_text="Version to upgrade to.",
    )

    class Meta:
        fields = ["version"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        instance = NetBoxConfiguration.objects.first()
        env = instance.get_env()

        versions = env.get_upgrades(include_patch=True)
        self.fields["version"].choices = [(str(version), str(version)) for version in versions]

    def clean(self):
        super().clean()

        if not get_workers_for_queue('default'):
            raise ValidationError(
                "No RQ workers operating on the 'default' queue are currently active in the environment."
            )

        instance = NetBoxConfiguration.objects.first()
        env = instance.get_env()

        # Run upgrade checks
        upgrade_check, error = env.upgrade_checks(self.cleaned_data.get("version"))
        if not upgrade_check:
            raise ValidationError({"version": error})

        # Ensure no actions are currently running on the environment
        if env.get_actions():
            raise ValidationError("There are currently actions running on the environment.")


class ConfirmationForm(_ConfirmationForm):
    """
    A generic confirmation form. The form is not valid unless the `confirm` field is checked.
    """

    name = forms.CharField(widget=forms.HiddenInput())

    def clean(self):
        super().clean()

        if not get_workers_for_queue('default'):
            raise ValidationError(
                "No RQ workers operating on the 'default' queue are currently active in the environment."
            )

        instance = NetBoxConfiguration.objects.first()
        env = instance.get_env()

        # Ensure no actions are currently running on the environment
        if env.get_actions():
            raise ValidationError("There are currently actions running on the environment.")
